import pytest
import os
import random
import time
from pathlib import Path

from dtaidistance import util_numpy, util
from dtaidistance.subsequence.dtw import subsequence_alignment, local_concurrences,\
    subsequence_search, LocalConcurrences
from dtaidistance import dtw_visualisation as dtwvis
from dtaidistance.exceptions import MatplotlibException
from dtaidistance.dtw import lb_keogh
from dtaidistance import dtw, dtw_ndim

directory = None
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")


@numpyonly
def test_dtw_subseq1():
    with util_numpy.test_uses_numpy() as np:
        query = np.array([1., 2, 0])
        series = np.array([1., 0, 1, 2, 1, 0, 2, 0, 3, 0, 0])
        sa = subsequence_alignment(query, series)
        mf = sa.matching_function()
        # print(f'{mf=}')
        match = sa.best_match()
        # print(match)
        # print(f'Segment={match.segment}')
        # print(f'Path={match.path}')
        if not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            if directory:
                plt.plot(mf)
                plt.savefig(directory / "subseq_matching.png")
                dtwvis.plot_warpingpaths(query, series, sa.warping_paths(), match.path,
                                         filename=directory / "subseq_warping.png")
                plt.close()
        best_k = sa.kbest_matches(k=3)
        assert match.path == [(0, 2), (1, 3), (2, 4)]
        assert [m.segment for m in best_k] == [[2, 4], [5, 7], [0, 1]]


@numpyonly
def test_dtw_subseq_eeg():
    with util_numpy.test_uses_numpy() as np:
        data_fn = Path(__file__).parent / 'rsrc' / 'EEGRat_10_1000.txt'
        data = np.loadtxt(data_fn)
        series = np.array(data[1500:1700])
        query = np.array(data[1331:1352])

        sa = subsequence_alignment(query, series)
        match = sa.best_match()
        kmatches = list(sa.kbest_matches(k=15, overlap=0))
        segments = [m.segment for m in kmatches]
        segments_sol = [[38, 56], [19, 37], [167, 185], [124, 143], [84, 100], [59, 77], [150, 162], [101, 121], [0, 15]]

        assert segments == segments_sol

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                raise MatplotlibException("No matplotlib available")

            fn = directory / "test_dtw_subseq_eeg1.png"
            fig = plt.figure(figsize=(20, 30))
            dtwvis.plot_warpingpaths(query, series, sa.warping_paths(), match.path, figure=fig)
            plt.savefig(fn)
            plt.close(fig)

            fn = directory / "test_dtw_subseq_eeg2.png"
            startidx, endidx = match.segment
            fig = plt.figure()
            plt.plot(query, label='query')
            plt.plot(series[startidx: endidx], label='best match')
            plt.legend()
            plt.savefig(fn)
            plt.close(fig)

            fn = directory / "test_dtw_subseq_eeg3.png"
            fig = plt.figure(figsize=(20, 10))
            fig, ax = dtwvis.plot_warpingpaths(query, series, sa.warping_paths(), path=-1, figure=fig)
            print('plotting {} matches'.format(len(kmatches)))
            for kmatch in kmatches:
                dtwvis.plot_warpingpaths_addpath(ax, kmatch.path)
            plt.savefig(fn)
            plt.close(fig)


@numpyonly
def test_dtw_subseq_bug1():
    use_c = True
    with util_numpy.test_uses_numpy() as np:
        query = np.array([-0.86271501, -1.32160597, -1.2307838, -0.97743775, -0.88183547,
                          -0.71453147, -0.70975136, -0.65238999, -0.48508599, -0.40860416,
                          -0.5567877, -0.39904393, -0.51854679, -0.51854679, -0.23652005,
                          -0.21261948, 0.16978966, 0.21281068, 0.6573613, 1.28355626,
                          1.88585065, 1.565583, 1.40305912, 1.64206483, 1.8667302])
        s1 = np.array([-0.87446789, 0.50009064, -1.43396157, 0.52081263, 1.28752619])
        s2 = np.array([1.19125347, 0.78778189, -0.95770272, -1.02133264])
        sa = subsequence_alignment(query, s1, use_c=use_c)
        assert sa.best_match().value == pytest.approx(0.08735692337954708)
        sa = subsequence_alignment(query, s2, use_c=use_c)
        assert sa.best_match().value == pytest.approx(0.25535859535443606)


@numpyonly
def test_dtw_subseq_ndim():
    use_c = False
    with util_numpy.test_uses_numpy() as np:
        # s1 = np.array([1., 2, 3,1])
        # query = np.array([2.0, 3.1])
        s1 = np.array([[1., 1], [2, 2], [3, 3], [1, 1]])
        query = np.array([[2.0, 2.1], [3.1, 3.0]])
        sa = subsequence_alignment(query, s1, use_c=use_c)
        m = sa.best_match()
        assert m.segment == [1, 2]
        assert m.value == pytest.approx(0.07071067811865482)


@numpyonly
@pytest.mark.parametrize("use_c", [False, True])
def test_dtw_subseq_ndim2(use_c):
    with util_numpy.test_uses_numpy() as np:
        s = [np.array([[1., 1], [2, 2], [3, 3]]),
             np.array([[2., 2], [3, 3], [1, 1]])]
        query = np.array([[2.0, 2.1], [3.1, 3.0]])
        d1 = [dtw_ndim.distance(si, query, use_c=use_c) for si in s]
        sa = subsequence_search(query, s, use_lb=False, use_c=use_c)
        assert str(sa.best_match()) == 'SSMatch(0)'
        d2 = [m.distance for m in sa.kbest_matches(k=2)]
        for d1i, d2i in zip(d1, d2):
            assert d1i == pytest.approx(d2i)


@pytest.mark.skip
@numpyonly
def test_dtw_localconcurrences_eeg():
    with util_numpy.test_uses_numpy() as np:
        data_fn = Path(__file__).parent / 'rsrc' / 'EEGRat_10_1000.txt'
        data = np.loadtxt(data_fn)
        series = np.array(data[1500:1700])

        gamma = 1
        # domain = 2 * np.std(series)
        # affinity = np.exp(-gamma * series)
        # print(f'Affinity in [{np.min(affinity)}, {np.max(affinity)}]\n'
        #       f'             {np.mean(affinity)} +- {np.std(affinity)}\n'
        #       f'             {np.exp(-gamma * np.mean(series))} +- {np.exp(-gamma * np.std(series))}\n'
        #       f'             {np.exp(-gamma * np.percentile(series, 75))} / {np.exp(-gamma * np.median(series))} / {np.exp(-gamma * np.percentile(series, 25))}\n')
        tau_stddev = 0.40
        diffp = tau_stddev*np.std(series)
        delta = -2 * np.exp(-gamma * diffp**2)  # -len(series)/2  # penalty
        delta_factor = 0.5
        tau = np.exp(-gamma * diffp**2)  # threshold
        # print(f'{tau=}, {delta=}')
        # tau=0.8532234738897421, delta=-1.7064469477794841
        buffer = -10
        minlen = 20
        tp = time.perf_counter()
        lc = local_concurrences(series, gamma=gamma, tau=tau, delta=delta, delta_factor=delta_factor,
                                use_c=True)
        tn = time.perf_counter()
        print(f'Align took: {tn-tp} seconds')
        print('{}, {}'.format(lc.tau, lc.delta))
        # print(f'{lc.tau=}, {lc.delta=}')
        matches = []
        tp = time.perf_counter()
        matches = lc.kbest_matches_store(k=100, minlen=minlen, buffer=buffer)
        tn = time.perf_counter()
        print(f'KBest matches took: {tn - tp} seconds')
        print( [(m.row, m.col) for m in matches])
        # assert [(m.row, m.col) for m in matches] == [(84, 95), (65, 93), (50, 117), (117, 200), (32, 180),
        #                                              (160, 178), (96, 139), (138, 181), (71, 200), (71, 117),
        #                                              (73, 137), (52, 138), (12, 117), (117, 178), (117, 160),
        #                                              (30, 160), (32, 52), (30, 117), (117, 135), (160, 200),
        #                                              (178, 200), (11, 52), (71, 160), (134, 160), (135, 200),
        #                                              (30, 200), (50, 200), (11, 73), (50, 160), (12, 33), (11, 137),
        #                                              (36, 143), (11, 179), (88, 160), (66, 178), (11, 93)]

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            fn = directory / "test_dtw_localconcurrences.png"
            fig = plt.figure()
            fig, ax = dtwvis.plot_warpingpaths(series, series, lc.wp_slice(), path=-1, figure=fig)
            for match in matches:
                dtwvis.plot_warpingpaths_addpath(ax, match.path)
            plt.savefig(fn)
            plt.close(fig)


@pytest.mark.skip
@numpyonly
def test_dtw_localconcurrences_short():
    with util_numpy.test_uses_numpy() as np:
        series1 = np.array([0., -1, -1, 0, 1, 2, 1, 0, 0, 0, 1, 3, 2, 1, 0, 0, 0, -1, 0])
        # series1 = np.array([0., -1, -1, 0, 1, 2, 1])
        # series2 = series1
        series2 = np.array([0.4, -0.9, -1.3, 1, 0.1, 2, 1, 0, 0, 10, 8, 3, 2, 1, 0, 0, 0, -1, 0])

        gamma = 1
        threshold_tau = 70
        delta = -2 * np.exp(-gamma * np.percentile(series1, threshold_tau))  # -len(series)/2  # penalty
        delta_factor = 0.1 #0.5
        tau = np.exp(-gamma * np.percentile(series1, threshold_tau))  # threshold
        # print(f'{tau=}, {delta=}, {delta_factor=}, {gamma=}')
        buffer = -10
        minlen = 3
        window = None
        lc = local_concurrences(series1, series2, gamma=gamma, tau=tau, delta=delta, delta_factor=delta_factor, penalty=1,
                                window=window, use_c=False)
        # print(lc.settings(kind="str"))
        # with np.printoptions(precision=2, linewidth=400):
        #     print(lc.wp)
        lc2 = local_concurrences(series1, series2, gamma=gamma, tau=tau, delta=delta, delta_factor=delta_factor, penalty=1,
                                 window=window, use_c=True, compact=True)
        # with np.printoptions(precision=2, linewidth=400):
        #     # print(lc2.wp)
        #     lc2.wp_c_print()
        np.testing.assert_allclose(lc.wp, lc2.wp_slice())
        p = lc.best_path(len(series1) - 1, len(series2))
        p2 = lc2.best_path(len(series1) - 1, len(series2))
        # print(p)
        # print(p2)
        assert str(p) == str(p2)
        # assert str(p) == "[(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9), (9, 10), (10, 11), "\
        #                  "(11, 12), (12, 13), (13, 14), (14, 15), (15, 16), (16, 17), (17, 18)]", \
        #                  str(p)
        matches = lc.kbest_matches_store(k=100, minlen=minlen, buffer=buffer)
        matches2 = lc2.kbest_matches_store(k=100, minlen=minlen, buffer=buffer)
        # print(matches)
        # print(matches2)
        # assert str(matches) == str(matches2)
        # print(lc2._wp)

        # assert [(m.row, m.col) for m in matches] == [(18, 19), (4, 19)]

        # print(lc2.wp_slice(2, 6, 3, 5))

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
                import matplotlib.backends.backend_pdf
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            fn = directory / "test_dtw_localconcurrences_short.pdf"
            pdf = matplotlib.backends.backend_pdf.PdfPages(fn)
            fig = plt.figure()
            fig, ax = dtwvis.plot_warpingpaths(series1, series2, lc.wp, path=-1, figure=fig)
            for match in matches:
                dtwvis.plot_warpingpaths_addpath(ax, match.path)
            pdf.savefig(fig)
            plt.close(fig)

            fig, ax = matches.plot(begin=None, end=None, showlegend=False)
            pdf.savefig(fig)
            plt.close(fig)

            fig, ax = matches2.plot(begin=None, end=None, showlegend=False)
            pdf.savefig(fig)
            plt.close(fig)
            pdf.close()


def create_data_subseqsearch_eeg(np, dtype=None, longer=False):
    window_extra = 200 if longer else 0
    data_fn = Path(__file__).parent / 'rsrc' / 'EEGRat_10_1000.txt'
    data = np.loadtxt(data_fn)
    if longer:
        series = np.array(data[:], dtype=dtype)
    else:
        series = np.array(data[1500:1700], dtype=dtype)
    query = np.array(data[1331:1352+window_extra], dtype=dtype)
    # print(f'{len(series)=}')

    k = 3
    s = []
    s_idx = []
    w = 22+window_extra  # window size
    ws = int(np.floor(w / 2))  # shift size
    wn = int(np.floor((len(series) - (w - ws)) / ws))
    si, ei = 0, w
    for i in range(wn):
        s.append(series[si:ei])
        s_idx.append(si)
        si += ws
        ei += ws
    return query, s, k, series, s_idx


@numpyonly
def test_dtw_subseqsearch_eeg2():
    with util_numpy.test_uses_numpy() as np:
        query, s, k, series, s_idx = create_data_subseqsearch_eeg(np)
        sa = subsequence_search(query, s, dists_options={'use_c': True})
        best = sa.kbest_matches_fast(k=k)
        assert str(best) == "[SSMatch(15), SSMatch(7), SSMatch(4)]", str(best)
        assert sa.distances is None

        sa = subsequence_search(query, s, dists_options={'use_c': True})
        best = sa.kbest_matches_fast(k=1)
        assert str(best) == "[SSMatch(15)]", str(best)

        sa = subsequence_search(query, s, dists_options={'use_c': True})
        best = sa.kbest_matches_fast(k=None)
        assert str(best) == "[SSMatch(15), SSMatch(7), SSMatch(4), SSMatch(11), SSMatch(6) ... SSMatch(14), SSMatch(10), SSMatch(9), SSMatch(1), SSMatch(3)]", str(best)

        assert best[0].value == pytest.approx(0.08045349583339727)

        sa = subsequence_search(query, s, dists_options={'use_c': True, 'max_dist': 0.0805 * len(query)})
        best = sa.kbest_matches_fast(k=k)
        assert str(best) == "[SSMatch(15)]", str(best)

        sa = subsequence_search(query, s, max_value=0.0805)
        best = sa.kbest_matches_fast(k=k)
        assert str(best) == "[SSMatch(15)]", str(best)

        sa = subsequence_search(query, s, max_dist=0.0805 * len(query))
        best = sa.kbest_matches_fast(k=k)
        assert str(best) == "[SSMatch(15)]", str(best)


@numpyonly
@pytest.mark.benchmark(group="subseqsearch_eeg")
def test_dtw_subseqsearch_eeg(benchmark):
    with util_numpy.test_uses_numpy() as np:
        query, s, k, series, s_idx = create_data_subseqsearch_eeg(np)

        def run():
            ss = subsequence_search(query, s, dists_options={'use_c': True})
            best = ss.kbest_matches_fast(k=k)
            return best, ss
        if benchmark is None:
            tic = time.perf_counter()
            best, sa = run()
            toc = time.perf_counter()
            print("Searching performed in {:0.4f} seconds".format(toc - tic))
        else:
            best, ss = benchmark(run)
        # print(sa.distances)
        # print(best)

        assert str(best) == "[SSMatch(15), SSMatch(7), SSMatch(4)]", str(best)
        assert str(best[0]) == str(ss.best_match()), '{} != {}'.format(best[0], ss.best_match())
        assert str(best[:]) == "[SSMatch(15), SSMatch(7), SSMatch(4)]", str(best[:])
        assert str(best[0:3]) == "[SSMatch(15), SSMatch(7), SSMatch(4)]", str(best[0:3])

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
                from matplotlib import gridspec
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            fn = directory / "test_dtw_subseqsearch_eeg.png"
            ymin, ymax = np.min(series), np.max(series)
            fig = plt.figure()
            if k is None:
                k = len(s)
            gs = gridspec.GridSpec(3, k, wspace=0.5, hspace=1)
            ax = fig.add_subplot(gs[0, 0])
            ax.plot(query)
            ax.set_title('Query')
            ax.set_ylim((ymin, ymax))
            for idx, match in enumerate(best):
                ax = fig.add_subplot(gs[1, idx])
                if idx == 0:
                    ax.set_title('Best {} windows'.format(k))
                ax.set_ylim((ymin, ymax))
                ax.plot(s[match.idx])
            ax = fig.add_subplot(gs[2, :])
            ax.set_ylim((ymin, ymax))
            ax.set_title('Series with windows')
            for idx in s_idx:
                ax.vlines(idx, ymin, ymax, color='grey', alpha=0.4)
            ax.plot(series)
            for idx, match in enumerate(best):
                ax.vlines(s_idx[match.idx], ymin, ymax, color='red')
            plt.savefig(fn)
            plt.close(fig)


@numpyonly
def test_lc_pat1():
    with util_numpy.test_uses_numpy() as np:
        data_fn = Path(__file__).parent / 'rsrc' / 'pat1.txt'
        data = np.loadtxt(data_fn)
        series1 = data[0, :]
        series2 = data[1, :]

        # gamma = 1
        # threshold_tau = 70
        # delta = -2 * np.exp(-gamma * np.percentile(series, threshold_tau))  # -len(series)/2  # penalty
        # delta_factor = 0.5
        # tau = np.exp(-gamma * np.percentile(series, threshold_tau))  # threshold
        # # print(f'{tau=}, {delta=}')
        buffer = 50
        minlen = 20
        t1 = time.time()
        lc = LocalConcurrences(series1, series2, window=int(buffer / 2), use_c=False)
        lc.estimate_settings(series1, series2, tau_type='mean', tau_factor=2)
        # lc.delta_factor = 1
        lc.exp_avg = None # 0.1
        lc.align()
        t2 = time.time()
        print(f'Time to align: {t2-t1}')

        print(lc.settings(kind="str"))
        t1 = time.time()
        matches = []
        for match in lc.kbest_matches(k=None, minlen=minlen, buffer=-buffer):
            if match is None:
                break
            matches.append(match)
        t2 = time.time()
        print(f'Time to find matches: {t2 - t1}')

        sm = lc.similarity_matrix()

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
                import matplotlib.backends.backend_pdf
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            fn = directory / "test_lc_pat1.pdf"
            pdf = matplotlib.backends.backend_pdf.PdfPages(fn)
            fig = plt.figure(figsize=(10, 10))
            fig, ax = dtwvis.plot_warpingpaths(series1, series2, sm,
                                               path=-1, figure=fig, showlegend=True,
                                               matshow_kwargs=lc.similarity_matrix_matshow_kwargs(sm))
            pdf.savefig(fig)

            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 10))
            values = sm  # np.array(sm.reshape((1, -1)))
            # print(np.histogram(values))
            ax.hist(values, bins=100, histtype='step', range=(-2, 1))
            ax.vlines([lc.tau, lc.delta], 0, 100, color='red')
            pdf.savefig(fig)

            fig = plt.figure(figsize=(10,10))
            fig, ax = dtwvis.plot_warpingpaths(series1, series2,
                                               lc.wp_slice(),
                                               path=-1, figure=fig, showlegend=True)
            for match in matches:
                dtwvis.plot_warpingpaths_addpath(ax, match.path)
            pdf.savefig(fig)

            fig = plt.figure(figsize=(10, 10))
            plt_slice = [180, 260]
            wp = lc.wp_slice(*plt_slice, *plt_slice)
            idx = wp < 0
            wp[idx] = -wp[idx]
            fig, ax = dtwvis.plot_warpingpaths(series1[slice(*plt_slice)], series2[slice(*plt_slice)],
                                               lc.wp_slice(*plt_slice, *plt_slice),
                                               path=-1, figure=fig, showlegend=True)
            for match in matches:
                dtwvis.plot_warpingpaths_addpath(ax, dtwvis.path_slice(match.path, *plt_slice, *plt_slice))
            pdf.savefig(fig)

            pdf.close()


@numpyonly
def test_lc_pat2():
    with util_numpy.test_uses_numpy() as np:
        series1 = np.array([75.01651784, 74.44948843, 73.38243493, 77.35680206,
               69.52647053, 63.21939026, 56.95414623, 49.94568125,
               45.33157425, 40.95059044, 41.90307336, 47.06055967,
               51.81505341, 55.43155354, 51.96320753, 48.20277142,
               45.00462966, 35.55600635, 32.1914493, 33.54732185,
               34.04776747, 35.40227115, 37.24460185, 32.33497074,
               31.56839457, 38.15810559, 39.529532, 29.82815339,
               19.36802916, 14.66462762, 13.16508061, 10.69378841,
               7.1061679, 6.12218125, 8.46480483, 11.15576087,
               11.91750573, 53.74608065, 135.66392315, 217.12666884,
               239.27151034, 210.42673095, 230.90243747, 256.56928804,
               243.7451023, 228.93578139, 224.0126408, 203.47771786,
               188.99680514, 194.56795499, 170.74918008, 138.98955878,
               124.02149478, 117.15055405, 86.49275773, 54.06789222,
               50.06429329, 47.90493883, 48.49793216, 45.46054247,
               36.37510044, 34.08652235, 27.73445741, 20.35611805,
               20.06074597, 20.66984978, 24.55085633, 29.0051147,
               30.67045557, 30.4834716])
        series2 = np.array([73.36588064,  69.89075092,  70.72715301,  71.09805833,
                            64.71067346,  62.81497759,  62.0549533 ,  56.25529614,
                            46.65040363,  41.25971095,  43.61368296,  45.47837141,
                            47.76064341,  50.8908327 ,  53.30984479,  52.99279867,
                            46.0746253 ,  38.24987477,  33.40804685,  33.12954541,
                            34.22982437,  33.82184342,  35.43014094,  32.7833301 ,
                            30.38732958,  34.17202522,  35.76355013,  29.71998153,
                            19.8391811 ,  14.60766209,  13.29786615,  10.73641285,
                            7.55388596,   6.67208187,   8.39639018,   9.95173822,
                            12.63007853,  28.26870444,  79.11514597, 141.86765425,
                            152.2452013 , 141.92643704, 161.33045751, 176.69692497,
                            164.85278629, 153.73041258, 154.77219994, 140.065507  ,
                            121.55208147, 111.32788681,  96.9295085 ,  73.39609933,
                            62.69909111,  59.72158109,  42.21259683,  30.58299875,
                            27.85303729,  27.50338718,  27.51806518,  24.46138111,
                            18.28259934,  19.72762923,  21.058755  ,  19.06932742,
                            21.35947407,  21.32860493,  23.90240201,  29.79843967,
                            33.49106013,  31.54302726])
        buffer = 50
        minlen = 20
        t1 = time.time()
        lc = LocalConcurrences(series1, series2, window=int(buffer / 2), use_c=False)
        lc.estimate_settings(series1, series2, tau_type='mean', tau_factor=2)
        # lc.delta_factor = 1
        lc.exp_avg = None  # 0.1
        lc.squash = 10
        lc.align()
        t2 = time.time()
        print(f'Time to align: {t2 - t1}')

        print(lc.settings(kind="str"))
        t1 = time.time()
        matches = []
        for match in lc.kbest_matches(k=None, minlen=minlen, buffer=-buffer):
            if match is None:
                break
            matches.append(match)
        t2 = time.time()
        print(f'Time to find matches: {t2 - t1}')

        sm = lc.similarity_matrix()

        if directory and not dtwvis.test_without_visualization():
            try:
                import matplotlib.pyplot as plt
                import matplotlib.backends.backend_pdf
            except ImportError:
                raise MatplotlibException("No matplotlib available")
            fn = directory / "test_lc_pat2.pdf"
            pdf = matplotlib.backends.backend_pdf.PdfPages(fn)
            fig = plt.figure(figsize=(10, 10))
            fig, ax = dtwvis.plot_warpingpaths(series1, series2, sm,
                                               path=-1, figure=fig, showlegend=True,
                                               matshow_kwargs=lc.similarity_matrix_matshow_kwargs(sm))
            pdf.savefig(fig)

            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 10))
            values = sm  # np.array(sm.reshape((1, -1)))
            # print(np.histogram(values))
            ax.hist(values, bins=100, histtype='step', range=(-2, 1))
            ax.vlines([lc.tau, lc.delta], 0, 100, color='red')
            pdf.savefig(fig)

            fig = plt.figure(figsize=(10, 10))
            print(lc.wp_slice(positivize=True))
            fig, ax = dtwvis.plot_warpingpaths(series1, series2,
                                               lc.wp_slice(positivize=True),
                                               path=-1, figure=fig, showlegend=True)
            for match in matches:
                dtwvis.plot_warpingpaths_addpath(ax, match.path)
            pdf.savefig(fig)

            # fig = plt.figure(figsize=(10, 10))
            # plt_slice = [180, 260]
            # fig, ax = dtwvis.plot_warpingpaths(series1[slice(*plt_slice)], series2[slice(*plt_slice)],
            #                                    lc.wp_slice(*plt_slice, *plt_slice),
            #                                    path=-1, figure=fig, showlegend=True)
            # for match in matches:
            #     dtwvis.plot_warpingpaths_addpath(ax, dtwvis.path_slice(match.path, *plt_slice, *plt_slice))
            # pdf.savefig(fig)

            pdf.close()


@numpyonly
@pytest.mark.benchmark(group="subseqsearch_eeg")
@pytest.mark.parametrize("use_c,use_lb", [(False, False), (True, False), (False, True), (True, True)])
def test_dtw_subseqsearch_eeg_lb(benchmark, use_c, use_lb):
    with util_numpy.test_uses_numpy() as np:
        query, s, k, series, s_idx = create_data_subseqsearch_eeg(np, longer=True)
        k = 1

        def run():
            sa = subsequence_search(query, s, use_c=use_c, use_lb=use_lb)
            best = sa.kbest_matches_fast(k=k)
            return best
        if benchmark is None:
            tic = time.perf_counter()
            best = run()
            toc = time.perf_counter()
            print("Searching performed in {:0.4f} seconds".format(toc - tic))
        else:
            best = benchmark(run)
        # print(sa.distances)
        # print(best)


@numpyonly
@pytest.mark.benchmark(group="test_eeg_lb")
@pytest.mark.parametrize("use_c", [False, True])
def test_eeg_lb(benchmark, use_c):
    with util_numpy.test_uses_numpy() as np:
        query, s, k, series, s_idx = create_data_subseqsearch_eeg(np, longer=False)
        k = 1

        def run():
            lb = []
            for serie in s:
                lb.append(lb_keogh(query, serie, use_c=use_c))
            return lb
        if benchmark is None:
            tic = time.perf_counter()
            lb = run()
            toc = time.perf_counter()
            print("Lowerbound performed in {:0.4f} seconds: {}".format(toc - tic, lb))
        else:
            best = benchmark(run)
        # print(sa.distances)
        # print(best)


@numpyonly
@pytest.mark.parametrize("use_c", [False, True])
def test_lb1(use_c):
    with util_numpy.test_uses_numpy() as np:
        a = np.array([1., 2, 1, 3])
        b = np.array([3., 4, 3, 0])
        lb = lb_keogh(a, b, window=2, use_c=use_c)
        assert lb == pytest.approx(2.23606797749979)


if __name__ == "__main__":
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print("Saving files to {}".format(directory))
    with util_numpy.test_uses_numpy() as np:
        np.set_printoptions(precision=6, linewidth=120)
        # test_dtw_subseq1()
        # test_dtw_subseq_eeg()
        # test_dtw_subseq_bug1()
        # test_dtw_subseq_ndim()
        # test_dtw_subseq_ndim2(use_c=True)
        # test_dtw_localconcurrences_eeg()
        # test_dtw_subseqsearch_eeg2()
        # test_lc_pat1()
        # test_lc_pat2()
        # import cProfile
        # cProfile.run('test_lc_pat1()')
        # test_dtw_subseqsearch_eeg2()
        # test_dtw_subseqsearch_eeg(benchmark=None)
        # test_dtw_subseqsearch_eeg_lb(benchmark=None, use_c=True, use_lb=False)
        # test_eeg_lb(benchmark=None, use_c=False)
        # test_dtw_localconcurrences_short()
        test_lb1(use_c=False)



# -*- coding: UTF-8 -*-
"""
dtaidistance.alignment
~~~~~~~~~~~~~~~~~~~~~~

Sequence alignment (e.g. Needleman–Wunsch).

:author: Wannes Meert
:copyright: Copyright 2017-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from .dp import dp, Direction

try:
    import numpy as np
except ImportError:
    np = None


def needleman_wunsch(s1, s2, window=None, max_dist=None,
                     max_step=None, max_length_diff=None, psi=None,
                     substitution=None):
    """Needleman-Wunsch global sequence alignment.

    Needleman-Wunsch finds the optimal aligment between two sequences by maximizing similarity.

    For background on this algorithm, see the original paper:
    Needleman, Saul B. & Wunsch, Christian D. (1970).
    "A general method applicable to the search for similarities in the amino acid sequence of two proteins".
    Journal of Molecular Biology. 48 (3): 443–53. doi:10.1016/0022-2836(70)90057-4 .

    It is equivalent to the Levenshtein distance that tries to minimize the edit distance. For more information see:
    Sellers PH (1974). "On the theory and computation of evolutionary distances".
    SIAM Journal on Applied Mathematics. 26 (4): 787–793. doi:10.1137/0126070 .

    :param s1: First sequence
    :param s2: Second sequence
    :param window:
    :param max_dist: Stop warping path if distance exceeds this value
    :param max_step: Stop warping path if the last increase exceeds this value.
    :param max_length_diff: Exit if the two sequences differ in length.
    :param psi: Psi relaxation
    :param substitution: Function with signature (s1[i], s2[j]) that returns the value to add to the
        current warping path given two symbols. For Needleman-Wunsch the default function that is used
         is -1 for a match and +1 for a mismatch.
         If you have custom distance values between particular symbols, you can use the `make_substitution_fn`
         method to generate a function from a dictionary.

    Example:

        >> s1 = "GATTACA"
        >> s2 = "GCATGCU"
        >> value, matrix = alignment.needleman_wunsch(s1, s2)
        >> algn, s1a, s2a = alignment.best_alignment(matrix, s1, s2)
        >> print(matrix)
           [[-0., -1., -2., -3., -4., -5., -6., -7.],
            [-1.,  1., -0., -1., -2., -3., -4., -5.],
            [-2., -0., -0.,  1., -0., -1., -2., -3.],
            [-3., -1., -1., -0.,  2.,  1., -0., -1.],
            [-4., -2., -2., -1.,  1.,  1., -0., -1.],
            [-5., -3., -3., -1., -0., -0., -0., -1.],
            [-6., -4., -2., -2., -1., -1.,  1., -0.],
            [-7., -5., -3., -1., -2., -2., -0., -0.]]
        >> print(''.join(s1a), ''.join(s2a))
            'G-ATTACA', 'GCAT-GCU'

    """
    if substitution is None:
        substitution = _default_substitution_fn
    value, scores, paths = dp(s1, s2,
                       fn=substitution, border=_needleman_wunsch_border,
                       penalty=0, window=window, max_dist=max_dist,
                       max_step=max_step, max_length_diff=max_length_diff, psi=psi)
    return -value, -scores, paths


def _needleman_wunsch_border(ri, ci):
    if ri == 0:
        return ci
    if ci == 0:
        return ri
    return 0


def _default_substitution_fn(v1, v2):
    """Default substitution function.

    Match: +1 -> -1
    Mismatch or Indel: −1 -> +1

    The values are reversed because our general dynamic programming algorithm
    selects the minimal value instead of the maximal value.
    """
    d_indel = 1  # gap / indel
    if v1 == v2:
        d = -1  # match
    else:
        d = 1  # mismatch
    return d, d_indel


def make_substitution_fn(matrix, gap=1, opt='max'):
    """Make a similarity function from a dictionary.
    
    Elements that are not in the dictionary are passed to the default
    function. This allows for this function to be used for only
    using the gap penalty as follows.

        substitution = make_substitution_fn({}, gap=0.5)

    :param matrix: Substitution matrix as a dictionary of tuples to values.
        For example `matrix={('A','B'): 2, ('B', 'A'): 3}`.
    :param opt: Direction in which matrix optimises alignments. If `max`,
        values are reversed, see :meth:` _default_substitution_fn`.
    :return: Function that compares two elements.
    """

    if opt == 'max':
        modifier = -1.0
    else:
        modifier = 1.0

    def _unwrap(a, b):
        if (a, b) in matrix:
            return matrix[(a, b)] * modifier, gap
        elif (b, a) in matrix:
            return matrix[(b, a)] * modifier, gap
        else:
            return _default_substitution_fn(a, b)[0], gap

    return _unwrap


def best_alignment(paths, s1=None, s2=None, gap="-", order=None):
    """Compute the optimal alignment from the nxm paths matrix.

    :param paths: Paths matrix (e.g. from needleman_wunsch)
    :param s1: First sequence, if given the aligned sequence will be created
    :param s2: Second sequence, if given the aligned sequence will be created
    :param gap: Gap symbol that is inserted into s1 and s2 to align the sequences
    :param order: Array with order of comparisons (there might be multiple optimal paths)
        The default order is 0,1,2: diagonal, up, left
        For example, 1,0,2 is up, diagonal, left
        There might be more optimal paths than covered by these orderings. For example,
        when using combinations of these orderings in different parts of the matrix.
    """
    i, j = int(paths.shape[0] - 1), int(paths.shape[1] - 1)
    p = [(i - 1, j - 1)]
    ops = [(-1,-1), (-1,-0), (-0,-1)]
    op_chars = [Direction.UP_LEFT.value, Direction.UP.value, Direction.LEFT.value]
    if order is None:
        order = [0, 1, 2]
    while i > 0 and j > 0:
        opi, opj = next(
            ops[orderi]
            for orderi in order
            if op_chars[orderi] in paths[i, j]
        )
        i, j = i + opi, j + opj
        p.append((i - 1, j - 1))
    while i > 0:
        i -= 1
        p.append((i -1, j - 1))
    while j > 0:
        j -= 1
        p.append((i -1, j - 1))

    s1a = None if s1 is None else []
    s2a = None if s2 is None else []
    s1ip, s2ip = p[0]
    for s1i, s2i in p[1:]:
        if s1i != s1ip and s2i != s2ip:
            # diagonal
            if s1a is not None:
                s1a.append(s1[s1ip])
            if s2a is not None:
                s2a.append(s2[s2ip])
        elif s1i == s1ip:
            if s1a is not None:
                s1a.append(gap)
            if s2a is not None:
                s2a.append(s2[s2ip])
        elif s2i == s2ip:
            if s1a is not None:
                s1a.append(s1[s1ip])
            if s2a is not None:
                s2a.append(gap)
        s1ip, s2ip = s1i, s2i
    if s1a is not None:
        s1a.reverse()
    if s2a is not None:
        s2a.reverse()

    p.pop()
    p.reverse()
    return p, s1a, s2a



import logging
import sys, os
import random
from pathlib import Path
sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))

import pytest

from dtaidistance.preprocessing import differencing
from dtaidistance import util_numpy

logger = logging.getLogger("be.kuleuven.dtai.distance")
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")
scipyonly = pytest.mark.skipif("util_numpy.test_without_scipy()")


@numpyonly
@scipyonly
def test_differencing():
    with util_numpy.test_uses_numpy() as np:
        series = np.array([0.1, 0.3, 0.2, 0.1] * 3)
        series = differencing(series, smooth=0.1)
        np.testing.assert_array_almost_equal(
            series, np.array([0.19982795,  0.10894791,  0.0429936 ,  0.00856575, -0.00471123, -0.00838414,
                              -0.00740218, -0.00785088, -0.02099783, -0.05340167, -0.09960599]))

        series = np.array([[0.1, 0.3, 0.2, 0.1] * 3])
        series = differencing(series, smooth=0.1)
        np.testing.assert_array_almost_equal(
            series, np.array([[0.19982795, 0.10894791, 0.0429936, 0.00856575, -0.00471123, -0.00838414,
                               -0.00740218, -0.00785088, -0.02099783, -0.05340167, -0.09960599]]))


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    test_differencing()



"""
dtaidistance.dtw_cc
~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW), C implementation.

:author: Wannes Meert
:copyright: Copyright 2017-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from cpython cimport array
import array
from cython import Py_ssize_t
from cython.view cimport array as cvarray
from libc.stdlib cimport abort, malloc, free, abs, labs
from libc.stdint cimport intptr_t
from libc.stdio cimport printf
from cpython.mem cimport PyMem_Malloc, PyMem_Free

cimport dtaidistancec_dtw
cimport dtaidistancec_globals
from dtaidistancec_dtw cimport seq_t


cdef extern from "Python.h":
    Py_ssize_t PY_SSIZE_T_MAX


cdef class DTWBlock:
    def __cinit__(self):
        pass

    def __init__(self, rb, re, cb, ce, triu=True):
        self._block.rb = rb
        self._block.re = re
        self._block.cb = cb
        self._block.ce = ce
        self._block.triu = triu

    @property
    def rb(self):
        return self._block.rb

    @property
    def re(self):
        return self._block.re

    def re_set(self, value):
        self._block.re = value

    @property
    def cb(self):
        return self._block.cb

    @property
    def ce(self):
        return self._block.ce

    def ce_set(self, value):
        self._block.ce = value

    @property
    def triu(self):
        return self._block.triu

    def triu_set(self, value):
        self._block.triu = value

    def __str__(self):
        return f'DTWBlock(rb={self.rb},re={self.re},cb={self.cb},ce={self.ce},triu={self.triu})'


cdef class DTWWps:
    def __cinit__(self):
        pass

    def __init__(self, l1, l2, DTWSettings settings):
        self._wps = dtaidistancec_dtw.dtw_wps_parts(l1, l2, &settings._settings)

    @property
    def ri1(self):
        return self._wps.ri1

    @property
    def ri2(self):
        return self._wps.ri2

    @property
    def ri3(self):
        return self._wps.ri3


cdef class DTWSettings:
    def __cinit__(self):
        pass

    def __init__(self, **kwargs):
        self._settings = dtaidistancec_dtw.dtw_settings_default()
        if "window" in kwargs:
            if kwargs["window"] is None:
                self._settings.window = 0
            else:
                self._settings.window = kwargs["window"]
        if "max_dist" in kwargs:
            if kwargs["max_dist"] is None:
                self._settings.max_dist = 0
            else:
                self._settings.max_dist = kwargs["max_dist"]
        if "max_step" in kwargs:
            if kwargs["max_step"] is None:
                self._settings.max_step = 0
            else:
                self._settings.max_step = kwargs["max_step"]
        if "max_length_diff" in kwargs:
            if kwargs["max_length_diff"] is None:
                self._settings.max_length_diff = 0
            else:
                self._settings.max_length_diff = kwargs["max_length_diff"]
        if "penalty" in kwargs:
            if kwargs["penalty"] is None:
                self._settings.penalty = 0
            else:
                self._settings.penalty = kwargs["penalty"]
        if "psi" in kwargs:
            if kwargs["psi"] is None:
                self._settings.psi_1b = 0
                self._settings.psi_1e = 0
                self._settings.psi_2b = 0
                self._settings.psi_2e = 0
            else:
                if type(kwargs["psi"]) is int:
                    self._settings.psi_1b = kwargs["psi"]
                    self._settings.psi_1e = kwargs["psi"]
                    self._settings.psi_2b = kwargs["psi"]
                    self._settings.psi_2e = kwargs["psi"]
                elif type(kwargs["psi"]) is tuple or type(kwargs["psi"]) is list:
                    if len(kwargs["psi"]) != 4:
                        self._settings.psi_1b = 0
                        self._settings.psi_1e = 0
                        self._settings.psi_2b = 0
                        self._settings.psi_2e = 0
                    else:
                        self._settings.psi_1b = kwargs["psi"][0]
                        self._settings.psi_1e = kwargs["psi"][1]
                        self._settings.psi_2b = kwargs["psi"][2]
                        self._settings.psi_2e = kwargs["psi"][3]
        if "use_pruning" in kwargs:
            if kwargs["use_pruning"] is None:
                self._settings.use_pruning = False
            else:
                self._settings.use_pruning = kwargs["use_pruning"]
        if "only_ub" in kwargs:
            if kwargs["only_ub"] is None:
                self._settings.only_ub = False
            else:
                self._settings.only_ub = kwargs["only_ub"]
        if "inner_dist" in kwargs:
            if kwargs["inner_dist"] == "squared euclidean" or kwargs["inner_dist"] == 0:
                self._settings.inner_dist = 0
            elif kwargs["inner_dist"] == "euclidean" or kwargs["inner_dist"] == 1:
                self._settings.inner_dist = 1
            else:
                raise AttributeError("Unknown inner_dist: {}".format(kwargs["inner_dist"]))

    @property
    def window(self):
        return self._settings.window

    @property
    def max_dist(self):
        return self._settings.max_dist

    @property
    def max_step(self):
        return self._settings.max_step

    @property
    def max_length_diff(self):
        return self._settings.max_length_diff

    @property
    def penalty(self):
        return self._settings.penalty

    @property
    def psi(self):
        return {
            '1b': self._settings.psi_1b,
            '1e': self._settings.psi_1e,
            '2b': self._settings.psi_2b,
            '2e': self._settings.psi_2e,
        }

    @property
    def use_pruning(self):
        return self._settings.use_pruning

    @property
    def only_ub(self):
        return self._settings.only_ub

    @property
    def inner_dist(self):
        if self._settings.inner_dist == 0:
            return "squared euclidean"
        elif self._settings.inner_dist == 1:
            return "euclidean"
        else:
            return "unknown inner distance"

    def __str__(self):
        return (
            "DTWSettings {\n"
            f"  window = {self.window}\n"
            f"  max_dist = {self.max_dist}\n"
            f"  max_step = {self.max_step}\n"
            f"  max_length_diff = {self.max_length_diff}\n"
            f"  penalty = {self.penalty}\n"
            f"  psi = {self.psi}\n"
            f"  use_pruning = {self.use_pruning}\n"
            f"  only_ub = {self.only_ub}\n"
            f"  inner_dist = {self.inner_dist}\n"
            "}")


cdef class DTWSeriesPointers:
    def __cinit__(self, int nb_series):
        self._ptrs = <seq_t **> malloc(nb_series * sizeof(seq_t*))
        self._nb_ptrs = nb_series
        if not self._ptrs:
            self._ptrs = NULL
            raise MemoryError()
        self._lengths = <Py_ssize_t *> malloc(nb_series * sizeof(Py_ssize_t))
        if not self._lengths:
            self._lengths = NULL
            raise MemoryError()

    def __dealloc__(self):
        if self._ptrs is not NULL:
            free(self._ptrs)
        if self._lengths is not NULL:
            free(self._lengths)


cdef class DTWSeriesMatrix:
    def __cinit__(self, seq_t[:, ::1] data):
        self._data = data

    @property
    def nb_rows(self):
        return self._data.shape[0]

    @property
    def nb_cols(self):
        return self._data.shape[1]


cdef class DTWSeriesMatrixNDim:
    def __cinit__(self, seq_t[:, :, ::1] data):
        self._data = data

    @property
    def nb_rows(self):
        return self._data.shape[0]

    @property
    def nb_cols(self):
        return self._data.shape[1]

    @property
    def nb_dims(self):
        return self._data.shape[2]


def dtw_series_from_data(data, force_pointers=False):
    cdef DTWSeriesPointers ptrs
    cdef DTWSeriesMatrix matrix
    cdef intptr_t ptr
    if force_pointers or isinstance(data, list) or isinstance(data, set) or isinstance(data, tuple):
        ptrs = DTWSeriesPointers(len(data))
        for i in range(len(data)):
            ptr = data[i].ctypes.data  # uniform for memoryviews and numpy
            ptrs._ptrs[i] = <seq_t *> ptr
            ptrs._lengths[i] = len(data[i])
        return ptrs
    try:
        matrix = DTWSeriesMatrix(data)
        return matrix
    except ValueError:
        pass
    try:
        matrix = DTWSeriesMatrixNDim(data)
        return matrix
    except ValueError:
        raise ValueError(f"Cannot convert data of type {type(data)}")


def ub_euclidean(seq_t[:] s1, seq_t[:] s2):
    """ See ed.euclidean_distance"""
    return dtaidistancec_dtw.ub_euclidean(&s1[0], len(s1), &s2[0], len(s2))


def ub_euclidean_ndim(seq_t[:, :] s1, seq_t[:, :] s2):
    """ See ed.euclidean_distance_ndim"""
    # Assumes C contiguous
    if s1.shape[1] != s2.shape[1]:
        raise Exception("Dimension of sequence entries needs to be the same: {} != {}".format(s1.shape[1], s2.shape[1]))
    ndim = s1.shape[1]
    return dtaidistancec_dtw.ub_euclidean_ndim(&s1[0,0], len(s1), &s2[0,0], len(s2), ndim)


def lb_keogh(seq_t[:] s1, seq_t[:] s2, **kwargs):
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    return dtaidistancec_dtw.lb_keogh(&s1[0], len(s1), &s2[0], len(s2), &settings._settings)


def distance(seq_t[:] s1, seq_t[:] s2, **kwargs):
    """DTW distance.

    Assumes C-contiguous arrays.

    See distance().
    :param s1: First sequence (buffer of seq_t-s)
    :param s2: Second sequence (buffer of seq_t-s)
    :param kwargs: Settings (see DTWSettings)
    """
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    return dtaidistancec_dtw.dtw_distance(&s1[0], len(s1), &s2[0], len(s2), &settings._settings)


def distance_ndim(seq_t[:, :] s1, seq_t[:, :] s2, **kwargs):
    """DTW distance for n-dimensional arrays.

    Assumes C-contiguous arrays.

    See distance().
    :param s1: First sequence (buffer of seq_t-s)
    :param s2: Second sequence (buffer of seq_t-s)
    :param ndim: Number of dimensions
    :param kwargs: Settings (see DTWSettings)
    """
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    if s1.shape[1] != s2.shape[1]:
        raise Exception("Dimension of sequence entries needs to be the same: {} != {}".format(s1.shape[1], s2.shape[1]))
    ndim = s1.shape[1]
    return dtaidistancec_dtw.dtw_distance_ndim(&s1[0,0], len(s1), &s2[0,0], len(s2), ndim, &settings._settings)


def distance_ndim_assinglearray(seq_t[:] s1, seq_t[:] s2, int ndim, **kwargs):
    """DTW distance for n-dimensional arrays.

    Assumes C-contiguous arrays (with sequence item as first dimension).

    See distance().
    :param s1: First sequence (buffer of seq_ts)
    :param s2: Second sequence (buffer of seq_ts)
    :param ndim: Number of dimensions
    :param kwargs: Settings (see DTWSettings)
    """
    # Assumes C contiguous
    settings = DTWSettings(**kwargs)
    return dtaidistancec_dtw.dtw_distance_ndim(&s1[0], len(s1), &s2[0], len(s2), ndim, &settings._settings)


def wps_length(Py_ssize_t l1, Py_ssize_t l2, **kwargs):
    settings = DTWSettings(**kwargs)
    return dtaidistancec_dtw.dtw_settings_wps_length(l1, l2, &settings._settings)


def wps_width(Py_ssize_t l1, Py_ssize_t l2, **kwargs):
    settings = DTWSettings(**kwargs)
    return dtaidistancec_dtw.dtw_settings_wps_width(l1, l2, &settings._settings)

{% set suffix = '' %}
{%- include 'dtw_cc_warpingpaths.jinja.pyx' %}

{% set suffix = '_ndim' %}
{%- include 'dtw_cc_warpingpaths.jinja.pyx' %}

{% set suffix = '_affinity' %}
{%- include 'dtw_cc_warpingpaths.jinja.pyx' %}

{% set suffix = '' %}
{%- include 'dtw_cc_warpingpath.jinja.pyx' %}

{% set suffix = '_ndim' %}
{%- include 'dtw_cc_warpingpath.jinja.pyx' %}

def wps_negativize_value(DTWWps p, seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c):
    dtaidistancec_dtw.dtw_wps_negativize_value(&p._wps, &wps[0,0], l1, l2, r, c)

def wps_positivize_value(DTWWps p, seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t r, Py_ssize_t c):
    dtaidistancec_dtw.dtw_wps_positivize_value(&p._wps, &wps[0,0], l1, l2, r, c)

def wps_negativize(DTWWps p, seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce):
    dtaidistancec_dtw.dtw_wps_negativize(&p._wps, &wps[0,0], l1, l2, rb, re, cb, ce)

def wps_positivize(DTWWps p, seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce):
    dtaidistancec_dtw.dtw_wps_positivize(&p._wps, &wps[0,0], l1, l2, rb, re, cb, ce)

def wps_max(DTWWps p, seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2):
    cdef Py_ssize_t r, c
    result = dtaidistancec_dtw.dtw_wps_max(&p._wps, &wps[0, 0], &r, &c, l1, l2)
    return r, c

def wps_expand_slice(seq_t[:, :] wps, seq_t[:, :] slice, Py_ssize_t l1, Py_ssize_t l2,
                     Py_ssize_t rb, Py_ssize_t re, Py_ssize_t cb, Py_ssize_t ce,
                     DTWSettings settings):
    dtaidistancec_dtw.dtw_expand_wps_slice_affinity(&wps[0, 0], &slice[0, 0],
                                                    l1, l2, rb, re, cb, ce,
                                                    &settings._settings)

def wps_print(seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, **kwargs):
    settings = DTWSettings(**kwargs)
    dtaidistancec_dtw.dtw_print_wps(&wps[0,0], l1, l2, &settings._settings)

def wps_print_compact(seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, **kwargs):
    settings = DTWSettings(**kwargs)
    dtaidistancec_dtw.dtw_print_wps_compact(&wps[0,0], l1, l2, &settings._settings)

def best_path_compact(seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, **kwargs):
    cdef Py_ssize_t path_length;
    settings = DTWSettings(**kwargs)
    cdef Py_ssize_t *i1 = <Py_ssize_t *> PyMem_Malloc((l1 + l2) * sizeof(Py_ssize_t))
    if not i1:
        raise MemoryError()
    cdef Py_ssize_t *i2 = <Py_ssize_t *> PyMem_Malloc((l1 + l2) * sizeof(Py_ssize_t))
    if not i2:
        raise MemoryError()
    try:
        path_length = dtaidistancec_dtw.dtw_best_path(&wps[0,0], i1, i2, l1, l2,
                                                      &settings._settings)
        path = []
        for i in range(path_length):
            path.append((i1[i], i2[i]))
        path.reverse()
    finally:
        PyMem_Free(i1)
        PyMem_Free(i2)
    return path

def best_path_compact_affinity(seq_t[:, :] wps, Py_ssize_t l1, Py_ssize_t l2, Py_ssize_t rs, Py_ssize_t cs, **kwargs):
    cdef Py_ssize_t path_length;
    settings = DTWSettings(**kwargs)
    cdef Py_ssize_t *i1 = <Py_ssize_t *> PyMem_Malloc((l1 + l2) * sizeof(Py_ssize_t))
    if not i1:
        raise MemoryError()
    cdef Py_ssize_t *i2 = <Py_ssize_t *> PyMem_Malloc((l1 + l2) * sizeof(Py_ssize_t))
    if not i2:
        raise MemoryError()
    try:
        path_length = dtaidistancec_dtw.dtw_best_path_affinity(&wps[0,0], i1, i2, l1, l2,
                                                               rs, cs, &settings._settings)
        path = []
        for i in range(path_length):
            path.append((i1[i], i2[i]))
        path.reverse()
    finally:
        PyMem_Free(i1)
        PyMem_Free(i2)
    return path


def srand(unsigned int seed):
    dtaidistancec_dtw.dtw_srand(seed)

def warping_path_prob(seq_t[:] s1, seq_t[:] s2, seq_t avg, include_distance=False, **kwargs):
    # Assumes C contiguous
    cdef Py_ssize_t path_length;
    cdef seq_t dist;
    settings = DTWSettings(**kwargs)
    cdef Py_ssize_t *i1 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i1:
        raise MemoryError()
    cdef Py_ssize_t *i2 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i2:
        raise MemoryError()
    try:
        dist = dtaidistancec_dtw.dtw_warping_path_prob_ndim(&s1[0], len(s1), &s2[0], len(s2), i1, i2, &path_length,
                                                            avg, 1, &settings._settings)
        path = []
        for i in range(path_length):
            path.append((i1[i], i2[i]))
        path.reverse()
    finally:
        PyMem_Free(i1)
        PyMem_Free(i2)
    if include_distance:
        return path, dist
    return path

{% set suffix = '' %}
{%- include 'dtw_cc_distancematrix.jinja.pyx' %}

{% set suffix = '_ndim' %}
{%- include 'dtw_cc_distancematrix.jinja.pyx' %}


def distance_matrix_length(DTWBlock block, Py_ssize_t nb_series):
    cdef Py_ssize_t length
    length = dtaidistancec_dtw.dtw_distances_length(&block._block, nb_series, nb_series)
    return length

{% set suffix = '' %}
{%- include 'dtw_cc_dba.jinja.pyx' %}

{% set suffix = '_ndim' %}
{%- include 'dtw_cc_dba.jinja.pyx' %}


cdef dba_inner(cur, seq_t *c_ptr, Py_ssize_t c_len, unsigned char *mask_ptr, int nb_prob_samples, int ndim, DTWSettings settings):
    cdef seq_t *matrix_ptr;
    cdef DTWSeriesMatrix matrix
    cdef DTWSeriesMatrixNDim matrix_ndim
    cdef DTWSeriesPointers ptrs
    if isinstance(cur, DTWSeriesMatrix) or isinstance(cur, DTWSeriesPointers):
        pass
    elif cur.__class__.__name__ == "SeriesContainer":
        cur = cur.c_data_compat()
    else:
        cur = dtw_series_from_data(cur)

    if isinstance(cur, DTWSeriesPointers):
        ptrs = cur
        dtaidistancec_dtw.dtw_dba_ptrs(
            ptrs._ptrs, ptrs._nb_ptrs, ptrs._lengths,
            c_ptr, c_len, mask_ptr, nb_prob_samples, ndim, &settings._settings)
    elif isinstance(cur, DTWSeriesMatrix):
        matrix = cur
        matrix_ptr = &matrix._data[0, 0]
        dtaidistancec_dtw.dtw_dba_matrix(
            matrix_ptr, matrix.nb_rows, matrix.nb_cols,
            c_ptr, c_len, mask_ptr, nb_prob_samples, ndim, &settings._settings)
    elif isinstance(cur, DTWSeriesMatrixNDim):
        matrix_ndim = cur
        matrix_ptr = &matrix_ndim._data[0, 0, 0]
        dtaidistancec_dtw.dtw_dba_matrix(
            matrix_ptr, matrix_ndim.nb_rows, matrix_ndim.nb_cols,
            c_ptr, c_len, mask_ptr, nb_prob_samples, ndim, &settings._settings)
    else:
        raise ValueError("Series are not the expected type (DTWSeriesPointers, DTWSeriesMatrix "
                         "or DTWSeriesMatrixNDim): {}".format(type(cur)))



#!/usr/bin/env python3
# encoding: utf-8
"""
check_installation.py

Created by Wannes Meert
Copyright (c) 2022 KU Leuven. All rights reserved.
"""

import sys

try:
    import dtaidistance
except ImportError as exc:
    print("Cannot import dtaidistance")
    sys.exit(1)

print('Location of dtaidistance:')
print(dtaidistance)

is_complete = dtaidistance.dtw.try_import_c(True)
if not is_complete:
    sys.exit(1)

sys.exit(0)




{%- set s = " " %}
def warping_path{{suffix}}(
        {%- if "ndim" in suffix -%}
        seq_t[:, :] s1, seq_t[:, :] s2, int ndim=1,{{s}}
        {%- else -%}
        seq_t[:] s1, seq_t[:] s2,{{s}}
        {%- endif -%}
        include_distance=False, **kwargs):
    # Assumes C contiguous
    cdef Py_ssize_t path_length;
    cdef seq_t dist;
    settings = DTWSettings(**kwargs)
    cdef Py_ssize_t *i1 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i1:
        raise MemoryError()
    cdef Py_ssize_t *i2 = <Py_ssize_t *> PyMem_Malloc((len(s1) + len(s2)) * sizeof(Py_ssize_t))
    if not i2:
        raise MemoryError()
    try:
        {%- if "ndim" in suffix %}
        dist = dtaidistancec_dtw.dtw_warping_path_ndim(&s1[0, 0], len(s1), &s2[0, 0], len(s2), i1, i2, &path_length, ndim, &settings._settings)
        {%- else %}
        dist = dtaidistancec_dtw.dtw_warping_path(&s1[0], len(s1), &s2[0], len(s2), i1, i2, &path_length, &settings._settings)
        {%- endif %}
        path = []
        for i in range(path_length):
            path.append((i1[i], i2[i]))
        path.reverse()
    finally:
        PyMem_Free(i1)
        PyMem_Free(i2)
    if include_distance:
        return path, dist
    return path



# -*- coding: UTF-8 -*-
"""
dtaidistance.sktime
~~~~~~~~~~~~~~~~~~~

Connectors to use DTAIDistance with sktime

:author: Wannes Meert
:copyright: Copyright 2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from .. import dtw
import numpy as np


def dtw_distance(x: np.ndarray, y: np.ndarray, dtw_settings=None, **kwargs) -> float:
    """Function compatible with sklearn.

    Can be used like:
    ```
    from sktime.classification.distance_based import KNeighborsTimeSeriesClassifier
    from dtaidistance.connectors.sktime import dtw_distance
    KNeighborsTimeSeriesClassifier(distance=dtw_distance)
    ```
    """
    if dtw_settings is None:
        dtw_settings = {}
    # DTAIDistance expact a row of values, sktime expects a column
    x = x[:, 0]
    y = y[:, 0]
    d = dtw.distance(x, y, **dtw_settings)
    return d


def dtw_distance_create(**hmtw_settings):
    def dtw_distance_fn(x, y, **kwargs):
        return dtw_distance(x, y, hmtw_settings, **kwargs)
    return dtw_distance_fn



# -*- coding: UTF-8 -*-
"""
dtaidistance.subsequence.localconcurrences
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.3.0 or higher)

DTW-based subsequence matching.

:author: Wannes Meert
:copyright: Copyright 2021-2023 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import math
from functools import partial

from .. import dtw  # import warping_paths, warping_paths_fast, best_path, warping_paths_affinity, distance
from .. import dtw_ndim
from .. import util_numpy
from .. import util


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    import numpy.ma as ma
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    ma = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None


def local_concurrences(series1, series2=None, gamma=1, tau=0, delta=0, delta_factor=1, estimate_settings=None,
                       only_triu=None, penalty=None, window=None, use_c=False, compact=None):
    """Local concurrences, see LocalConcurrences.

    :param series1:
    :param series2:
    :param gamma: Affinity transformation exp(-gamma*(s1[i] - s2[j])**2)
    :param tau: threshold parameter
    :param delta: penalty parameter
        Should be negative. Added instead of the affinity score (if score below tau threshold parameter).
    :param delta_factor: multiply cumulative score (e.g. by 0.5).
        This is useful to have the same impact at different locations in the warping paths matrix, which
        is cumulative (and thus typically large in one corner and small in the opposite corner).
    :param estimate_settings: Estimate tau, delta, delta_factor from given series. Will be passed as
        tau_std to estimate_settings_from_std.
    :param only_triu: Only compute the upper traingle matrix values. Useful to avoid redundant computations
        when series1 is equal to series2 (or equivalently if series2 is None).
    :param penalty: Penalty that is added when dynamic programming is using moving vertically or horizontally
        through the matrix instead of diagonally. Used to prefer diagonal paths.
    :param compact: Use the compact representation for the warping paths matrix (only when use_c is true).
    :return:
    """
    lc = LocalConcurrences(series1, series2, gamma, tau, delta, delta_factor,
                           only_triu=only_triu, penalty=penalty, window=window, use_c=use_c, compact=compact)
    if estimate_settings is not None:
        lc.estimate_settings_from_std(series1, series2, tau_std=estimate_settings)
    lc.align()
    return lc


class LCMatch:
    def __init__(self, lc, row=None, col=None):
        """LocalConcurrences match"""
        self.row = row  # type: int
        self.col = col  # type: int
        self.lc = lc  # type: LocalConcurrences
        self._path = None

    @property
    def path(self):
        if self._path is not None:
            return self._path
        # TODO: always storing the path might be memory hungry
        #       but recomputing is impossible since the values are negated/masked afterwards
        self._path = self.lc.best_path(self.row, self.col)
        return self._path

    def distance(self, do_sqrt=True):
        if self._path is None:
            return None
        d = 0
        for r, c in self._path:
            d += (self.lc.series1[r] - self.lc.series2[c])**2
        if do_sqrt:
            d = math.sqrt(d)
        return d

    def __str__(self):
        return f'LCMatch({self.row, self.col})'

    def __repr__(self):
        return self.__str__()


class LCMatches:
    def __init__(self, lc,  matches=None):
        self._lc = lc
        self._matches = []
        if matches is not None:
            self._matches.update(matches)

    def __iter__(self):
        return self._matches.__iter__()

    def append(self, match):
        self._matches.append(match)

    def covered(self):
        s1 = np.zeros(len(self._lc.series1), dtype=np.bool)
        s2 = np.zeros(len(self._lc.series2), dtype=np.bool)
        for match in self._matches:
            path = match.path
            s1[path[0][0]:path[-1][0]+1] = True
            s2[path[0][1]:path[-1][1]+1] = True
        return s1, s2

    def coverage(self):
        s1, s2 = self.covered()
        c1 = np.sum(s1) / len(s1)
        c2 = np.sum(s2) / len(s2)
        return c1, c2

    def segments(self):
        s1, s2 = [], []
        for match in self._matches:
            p = match.path
            s1.append((p[0][0], p[-1][0]))
            s2.append((p[0][1], p[-1][1]))
        return s1, s2

    def missing(self):
        s1, s2 = self.coverage()
        s1 = ~s1
        s2 = ~s2
        return s1, s2

    def missing_segments(self):
        b1, b2 = self.covered()
        s1, s2 = self.segments()
        for sb, se in s1:
            b1[sb] = False
            b1[se] = False
        for sb, se in s2:
            b2[sb] = False
            b2[se] = False

        inmissing = False
        ms1 = []
        lstart = None
        for i in range(len(b1)):
            if inmissing:
                if b1[i]:
                    ms1.append((lstart, i))
                    inmissing = False
            else:
                if not b1[i]:
                    lstart = i
                    inmissing = True

        inmissing = False
        ms2 = []
        lstart = None
        for i in range(len(b2)):
            if inmissing:
                if b2[i]:
                    ms2.append((lstart, i))
                    inmissing = False
            else:
                if not b2[i]:
                    lstart = i
                    inmissing = True
        return ms1, ms2

    def distance(self, do_sqrt=True):
        d = 0
        for m in self._matches:
            d += m.distance(do_sqrt=False)
        if do_sqrt:
            d = math.sqrt(d)
        return d

    def distance_compensated(self, penalty=None, max_factor=10):
        """Distance with compensation for missed parts in sequences.

        :param penalty: Base penalty per missing step in the joint path
        :param max_factor: Number >1
        """
        if penalty is None:
            penalty = 1/self._lc.gamma
        d = self.distance(do_sqrt=False)
        c1, c2 = self.coverage()
        perc_missing = 1 - max(c1, c2)
        nb_missing = max((1-c1)*len(self._lc.series1), (1-c2)*len(self._lc.series2))
        if max_factor is not None:
            factor = 1 + ((max_factor - 1) * perc_missing)
        else:
            factor = 1
        d += factor * penalty * nb_missing
        d = math.sqrt(d)
        return d

    def plot(self, begin=None, end=None, showlegend=False, showpaths=True, showboundaries=True):
        from .. import dtw_visualisation as dtwvis
        if begin is None and end is None:
            series1 = self._lc.series1
            series2 = self._lc.series2
            wp = self._lc.wp_slice()
            begin = 0
        elif begin is None:
            series1 = self._lc.series1[:end]
            series2 = self._lc.series2[:end]
            wp = self._lc.wp_slice(re=end, ce=end)
            begin = 0
        elif end is None:
            series1 = self._lc.series1[begin:]
            series2 = self._lc.series2[begin:]
            wp = self._lc.wp_slice(rb=begin, cb=begin)
        else:
            series1 = self._lc.series1[begin:end]
            series2 = self._lc.series2[begin:end]
            wp = self._lc.wp_slice(rb=begin, re=end, cb=begin, ce=end)
        if begin is not None and begin > 0:
            includes_zero = False
        else:
            includes_zero = True
        fig, ax = dtwvis.plot_warpingpaths(series1, series2, wp, path=-1, showlegend=showlegend, includes_zero=includes_zero)
        if showpaths:
            nb_plotted = 0
            for i, match in enumerate(self._matches):
                path2 = []
                for t in match.path:
                    if begin is not None and (t[0] < begin or t[1] < begin):
                        continue
                    if end is not None and (t[0] > (end-1) or t[1] > (end-1)):
                        continue
                    path2.append((t[0] - begin, t[1] - begin))
                if len(path2) > 0:
                    nb_plotted += 1
                    dtwvis.plot_warpingpaths_addpath(ax, path2)
            print(f"Paths plotted: {nb_plotted}")
        if showboundaries:
            # s1, s2 = self.covered()
            ss1, ss2 = self.segments()
            sbs, ses = zip(*ss1)
            sbs1 = [v - begin for v in sbs if (begin is None or v >= begin) and (end is None or v <= end)]
            ses1 = [v - begin for v in ses if (begin is None or v >= begin) and (end is None or v <= end)]
            ax[3].hlines(sbs1, 0, len(series2) - 1, color='black', alpha=0.5)
            ax[3].hlines(ses1, 0, len(series2) - 1, color='black', alpha=0.5)
            sbs, ses = zip(*ss2)
            sbs2 = [v - begin for v in sbs if (begin is None or v >= begin) and (end is None or v <= end)]
            ses2 = [v - begin for v in ses if (begin is None or v >= begin) and (end is None or v <= end)]
            ax[3].vlines(sbs2, 0, len(series1) - 1, color='black', alpha=0.5)
            ax[3].vlines(ses2, 0, len(series1) - 1, color='black', alpha=0.5)
            ymin = min(np.min(series1), np.min(series2))
            for idx, (sb, se) in enumerate(zip(sbs1, ses1)):
                ax[2].plot([-ymin, -ymin], [len(series1)-sb, len(series1)-se], color='blue', linewidth=2, alpha=0.5)
            for idx, (sb, se) in enumerate(zip(sbs2, ses2)):
                ax[1].plot([sb, se], [ymin, ymin], color='blue', linewidth=2, alpha=0.5)
        return fig, ax

    def str(self, maxlength=10):
        return '[' + ', '.join(str(m) for m in self._matches[:maxlength]) + ']'

    def __str__(self):
        return self.str()


class LocalConcurrences:
    def __init__(self, series1, series2=None, gamma=1, tau=0, delta=0, delta_factor=1, only_triu=False,
                 penalty=None, window=None, use_c=False, compact=None):
        """Version identification based on local concurrences.

        Find recurring patterns across two time series. Used to identify whether one time series is
        a version of another. If the two time series are the same one, it can be used to find typical
        or frequent patterns in a time series.

        Based on 7.3.2 Identiﬁcation Procedure in Fundamentals of Music Processing, Meinard Müller, Springer, 2015.

        Different from the original formulation, D_tau is introduced based on the given delta factor.
        This makes the penalty less sensitive to the cumulative effect of the paths in the
        self-similarity matrix S:

        S_tau(n,m) = S(n,m)  if  S(n,m) >= tau  (with tau >= 0)
                     delta   if  S(n,m) < tau   (with tau >= 0 & delta <= 0)

        And for the accumulated score matrix D:

        D_tau(n,m) = max(0,
                         df * D_tau(n−1,m−1) + S_tau(n,m),
                         df * D_tau(n−1,m)   + S_tau(n,m),
                         df * D_tau(n,m−1)   + S_tau(n,m))
        where df = 1 if S(n,m) >= tau and df=delta_factor (<=1) otherwise,

        For finding paths the delta_factor has no influence. For the visualisation,
        it helps as patterns exhibit more similar values in the D matrix.

        :param series1: First time series.
        :param series2: Second time series. If empty, series1 is used and compared with itself.
        :param gamma: Affinity transformation exp(-gamma*(s1[i] - s2[j])**2), should be >0
        :param tau: threshold parameter, should be >= 0
        :param delta: penalty parameter, should be <= 0
        :param delta_factor: penalty factor parameter, should be <= 1
        :param only_triu: Only consider upper triangular matrix in warping paths.
        :param compact: Use the compact representation for the warping paths matrix (only when use_c is true).
        """
        self.series1 = series1
        if series2 is None:
            # Self-comparison
            self.series2 = self.series1
            self.only_triu = True if only_triu is None else only_triu
        else:
            self.series2 = series2
            self.only_triu = False if only_triu is None else only_triu
        self.gamma = gamma
        self.tau = tau
        self.delta = delta
        self.delta_factor = delta_factor
        self.penalty = penalty
        self.window = window
        self.use_c = use_c
        if compact is None:
            self.compact = self.use_c
        else:
            self.compact = compact
        self._wp = None  # warping paths
        if self.use_c:
            self._c_settings = dtw_cc.DTWSettings(window=self.window, penalty=self.penalty)
            self._c_parts = dtw_cc.DTWWps(len(self.series1), len(self.series2), self._c_settings)

    @staticmethod
    def from_other(lc, series1, series2=None):
        lcn = LocalConcurrences(series1, series2, gamma=lc.gamma, tau=lc.tau, delta=lc.delta,
                                delta_factor=lc.delta_factor, only_triu=lc.only_triu,
                                penalty=lc.penalty, window=lc.window, use_c=lc.use_c, compact=lc.compact)
        return lcn

    def reset(self):
        self._wp = None

    def estimate_settings_from_std(self, series, series2=None, tau_std=0.33):
        """Estimate delta, tau and delta_factor from series, tau_std and gamma.

        :param series:
        :param tau_std: Set tau to differences larger than tau_std time standard deviation of
            the given series (default is 0.33, or reject differences that are larger than
            the deviation wrt to the mean of 75% of the values in the series, assuming a
            normal distribution).
        :return:
        """
        return self.estimate_settings(series, series2, tau_type='std', tau_factor=tau_std)

    def estimate_settings_from_mean(self, series, series2=None, tau_mean=0.33):
        return self.estimate_settings(series, series2, tau_type='mean', tau_factor=tau_mean)

    def estimate_settings_from_abs(self, series, series2=None, tau_abs=0.33):
        return self.estimate_settings(series, series2, tau_type='abs', tau_factor=tau_abs)

    def estimate_settings(self, series, series2=None, tau_factor=0.33, tau_type='mean', gamma=None):
        if tau_type != 'abs':
            if series is None:
                diffm = 1
            elif series2 is None:
                if tau_type == 'std':
                    diffm = np.std(series)
                elif tau_type == 'mean':
                    diffm = np.mean(series)
                else:
                    diffm = 1
            else:
                if tau_type == 'std':
                    diffm = np.std(np.abs(series - series2))
                elif tau_type == 'mean':
                    diffm = np.mean(np.abs(series - series2))
                else:
                    diffm = 1

            if gamma is None:
                # Intuition for gamma:
                # Create an affinity matrix where
                # differences up to the mean/std are in [e^-1, 1],
                # larger differences are i [0, e^-1]
                self.gamma = 1 / diffm**2
            else:
                self.gamma = gamma
            if tau_factor is not None:
                diffp = tau_factor * diffm
            else:
                diffp = diffm
        elif tau_type == 'abs':
            diffp = tau_factor
        else:
            raise AttributeError('{} is not supported (not in mean, std, abs)'.format(tau_type))
        self.tau = np.exp(-self.gamma * diffp ** 2)
        self.delta = -2 * self.tau
        self.delta_factor = 0.90
        self.penalty = self.tau / 10

    def align(self):
        """

        :return:
        """
        if self._wp is not None:
            return
        if self.use_c:
            fn = partial(dtw.warping_paths_affinity_fast, compact=self.compact)
        else:
            fn = dtw.warping_paths_affinity
        _, wp = fn(self.series1, self.series2,
                   gamma=self.gamma, tau=self.tau, delta=self.delta, delta_factor=self.delta_factor,
                   only_triu=self.only_triu, penalty=self.penalty, window=self.window)
        if self.compact:
            self._wp = wp
        else:
            self._wp = ma.masked_array(wp)
        self._reset_wp_mask()
        # if self.only_triu:
        #     il = np.tril_indices(self._wp.shape[0])
        #     self._wp[il] = ma.masked

    def align_fast(self):
        use_c = self.use_c
        self.use_c = True
        result = self.align()
        self.use_c = use_c
        return result

    def _reset_wp_mask(self):
        if self.compact:
            dtw_cc.wps_positivize(self._c_parts, self._wp,
                                  len(self.series1), len(self.series2),
                                  0, len(self.series1) + 1,
                                  0, len(self.series2) + 1)
        else:
            wp = self._wp
            if self.window is None:
                wp.mask = False
            else:
                windowdiff1 = max(0, wp.shape[1] - wp.shape[0])
                windowdiff2 = max(0, wp.shape[0] - wp.shape[1])
                il = np.tril_indices(n=wp.shape[0], k=-1 - self.window - windowdiff2, m=wp.shape[1])
                wp[il] = ma.masked
                il = np.triu_indices(n=wp.shape[0], k=-self.window - windowdiff2, m=wp.shape[1])
                wp.mask[il] = False
                il = np.triu_indices(n=wp.shape[0], k=1 + self.window + windowdiff1, m=wp.shape[1])
                wp[il] = ma.masked
            if self.only_triu:
                il = np.tril_indices(self._wp.shape[0], k=-1)
                wp[il] = -np.inf
                wp[il] = ma.masked

    def similarity_matrix(self):
        sm = ma.masked_array(np.empty((len(self.series1), len(self.series2))))
        for r in range(len(self.series1)):
            if self.window is None:
                minc, maxc = 0, len(self.series2)
            else:
                minc, maxc = max(0, r - self.window), min(len(self.series2), r + self.window)
            for c in range(minc):
                sm[r, c] = ma.masked
            for c in range(minc, maxc):
                d = np.exp(-self.gamma * (self.series1[r] - self.series2[c]) ** 2)
                sm[r, c] = self.delta if d < self.tau else d
            for c in range(maxc, len(self.series2)):
                sm[r, c] = ma.masked
        return sm

    def similarity_matrix_matshow_kwargs(self, sm):
        import matplotlib.pyplot as plt
        from matplotlib.colors import BoundaryNorm
        # viridis = cm.get_cmap('viridis', 256)
        # newcolors = viridis(np.linspace(0, 1, 256))
        # pink = np.array([248 / 256, 24 / 256, 148 / 256, 1])
        # newcolors[:25, :] = pink
        # newcmp = ListedColormap(newcolors)
        # define the colormap
        cmap = plt.get_cmap('Spectral')
        # extract all colors from the .jet map
        cmaplist = [cmap(i) for i in range(cmap.N)]
        # create the new map
        cmap = cmap.from_list('Custom cmap', cmaplist, cmap.N)
        # define the bins and normalize and forcing 0 to be part of the colorbar!
        sm[sm == -np.inf] = 0
        sm_max = max(1, np.max(sm))
        sm_min = min(self.delta, np.min(sm))
        # bounds_pos = np.arange(0, sm_max, .01)
        bounds_pos = np.linspace(0, sm_max, 128)
        bounds_neg = np.linspace(sm_min, 0, len(bounds_pos))
        bounds = np.concatenate((bounds_neg, bounds_pos))
        # bounds = np.arange(sm_min, sm_max, .01)
        # idx = np.searchsorted(bounds, 0)
        # bounds = np.insert(bounds, idx, 0)
        norm = BoundaryNorm(bounds, cmap.N)
        return {'cmap': cmap, 'norm': norm}

    @property
    def wp(self):
        if self.compact:
            raise NotImplementedError("The full warping paths matrix is not available when using compact=True.\n"
                                      "Use wp_slice to construct part of the matrix from the compact data structure.")
        return self._wp.data

    def wp_slice(self, rb=None, re=None, cb=None, ce=None, positivize=False):
        if rb is None:
            rb = 0
        if re is None:
            re = len(self.series1) + 1
        if cb is None:
            cb = 0
        if ce is None:
            ce = len(self.series2) + 1
        if not (0 <= rb <= len(self.series1) + 1 and
                0 <= re <= len(self.series1) + 1 and
                0 <= cb <= len(self.series2) + 1 and
                0 <= ce <= len(self.series2) + 1):
            raise ValueError('Slice needs to be in 0<=r<={} and 0<=c<={}'.format(len(self.series1) + 1,
                                                                                 len(self.series2) + 1))
        if self.compact:
            slice = np.empty((re-rb, ce-cb), dtype=np.double)
            dtw_cc.wps_expand_slice(self._wp, slice, len(self.series1), len(self.series2),
                                    rb, re, cb, ce, self._c_settings)
        else:
            slice = self._wp[rb:re, cb:ce]
        if positivize:
            neg_idx = slice < 0
            slice[neg_idx] = -slice[neg_idx]
        return slice

    def best_match(self):
        idx = np.unravel_index(np.argmax(self._wp, axis=None), self._wp.shape)
        r, c = idx
        lcm = LCMatch(self, r, c)
        # path = lcm.path
        # for (x, y) in path:
        #     self._wp[x + 1, y + 1] = ma.masked
        return lcm

    def kbest_matches_store(self, k=1, minlen=2, buffer=0, restart=True, keep=False, matches=None, tqdm=None):
        import time
        if matches is None:
            matches = LCMatches(self)
        it = self.kbest_matches(k=k, minlen=minlen, buffer=buffer, restart=restart)
        if tqdm is not None:
            it = tqdm(it, total=k)
        tp = time.perf_counter()
        for ki, match in enumerate(it):
            matches.append(match)
            tn = time.perf_counter()
            #print(f'time: {tn-tp}')
            tp = tn
        if not keep:
            self._reset_wp_mask()
        return matches

    def kbest_matches(self, k=1, minlen=2, buffer=0, restart=True):
        """Yields the next best LocalConcurrent match.
        Stops at k matches (use None for all matches).

        :param k: Number of matches to yield, None is all matches
        :param minlen: Consider only matches of length longer than minlen
        :param buffer: Matches cannot be closer than buffer to each other
        :param restart: Start searching from start, ignore previous calls to kbest_matches
        :param keep: Keep mask to search incrementally for multiple calls of kbest_matches
        :return: Yield an LCMatch object
        """
        if self._wp is None:
            self.align()
        wp = self._wp
        if restart:
            self._reset_wp_mask()
        l1 = len(self.series1)
        l2 = len(self.series2)
        lperc = max(100, int(l1/10))
        ki = 0
        while k is None or ki < k:
            idx = None
            lcm = None
            cnt = 0
            while idx is None:
                cnt += 1
                if cnt % lperc == 0:
                    print(f'Searching for matches is taking a long time (k={ki+1}/{k}: {cnt} tries)')
                if self.compact:
                    idx = dtw_cc.wps_max(self._c_parts, wp, l1, l2)
                else:
                    idx = np.unravel_index(np.argmax(wp, axis=None), wp.shape)
                if idx[0] == 0 or idx[1] == 0:
                    # If all are masked, idx=0 is returned
                    return None
                r, c = idx
                # print(f'Best value: wp[{r},{c}] = {wp[r,c]}')
                lcm = LCMatch(self, r, c)
                path = lcm.path
                for (x, y) in path:
                    x += 1
                    y += 1
                    if not self.compact:
                        if len(wp.mask.shape) > 0 and wp.mask[x, y] is True:  # True means invalid
                            # print('found path contains masked, restart')
                            lcm = None
                            idx = None
                            break
                        else:
                            wp[x, y] = -wp[x, y]  # ma.masked
                    else:
                        dtw_cc.wps_negativize_value(self._c_parts, wp, l1, l2, x, y)
                if len(path) < minlen:
                    # print('found path too short, restart')
                    lcm = None
                    idx = None
            if buffer < 0 and lcm is not None:
                if self.compact:
                    dtw_cc.wps_negativize(self._c_parts, wp,
                                          len(self.series1), len(self.series2),
                                          path[0][0]+1, path[-1][0]+2,
                                          path[0][1]+1, path[-1][1]+2)
                else:
                    miny, maxy = 0, wp.shape[1]
                    minx, maxx = 0, wp.shape[0]
                    wp[path[0][0]+1:path[-1][0]+2, miny:maxy] = -wp[path[0][0]+1:path[-1][0]+2, miny:maxy]  # ma.masked
                    wp[minx:maxx, path[0][1]+1:path[-1][1]+2] = -wp[minx:maxx, path[0][1]+1:path[-1][1]+2]  # ma.masked
            elif buffer > 0 and lcm is not None:
                miny, maxy = 0, wp.shape[1] - 1
                minx, maxx = 0, wp.shape[0] - 1
                if self.compact:
                    raise Exception("A positive buffer is not yet supported for compact WP data structure")
                else:
                    for (x, y) in path:
                        xx = x + 1
                        for yy in range(max(miny, y + 1 - buffer), min(maxy, y + 1 + buffer)):
                            wp[xx, yy] = -wp[xx, yy]  # ma.masked
                        yy = y + 1
                        for xx in range(max(minx, x + 1 - buffer), min(maxx, x + 1 + buffer)):
                            wp[xx, yy] = -wp[xx, yy]  # ma.masked
            if lcm is not None:
                ki += 1
                yield lcm

    def best_path(self, row, col, wp=None):
        if self._wp is None:
            return None
        if wp is None:
            wp = self._wp
        l1 = len(self.series1)
        l2 = len(self.series2)
        if self.compact:
            p = dtw_cc.best_path_compact_affinity(wp, l1, l2, row, col, window=self.window)
            return p
        argm = argmax
        i = row
        j = col
        p = [(i - 1, j - 1)]
        # prev = self._wp[i, j]
        while i > 0 and j > 0:
            values = [wp[i - 1, j - 1], wp[i - 1, j], wp[i, j - 1]]
            # print(f'{i=}, {j=}, {argm(values)=}, {ma.argmax(values)=}, {values=}')
            values = [-1 if v is ma.masked else v for v in values]
            c = argmax(values)  # triggers "Warning: converting a masked element to nan"
            # if values[c] is ma.masked:
            #     break
            if values[c] <= 0:  # values[c] > prev:
                break
            # prev = values[c]
            if c == 0:
                if wp[i - 1, j - 1] is ma.masked or wp[i - 1, j - 1] < 0:
                    assert False
                    break
                i, j = i - 1, j - 1
            elif c == 1:
                if wp[i - 1, j] is ma.masked or wp[i - 1, j] < 0:
                    assert False
                    break
                i = i - 1
            elif c == 2:
                if wp[i, j - 1] is ma.masked or wp[i, j - 1] < 0:
                    assert False
                    break
                j = j - 1
            p.append((i - 1, j - 1))
        if p[-1][0] < 0 or p[-1][1] < 0:
            p.pop()
        p.reverse()
        return p

    def settings_from(self, lc):
        self.gamma = lc.gamma
        self.tau = lc.tau
        self.delta = lc.delta
        self.delta_factor = lc.delta_factor
        self.penalty = lc.penalty
        self.window = lc.window

    def settings(self, kind=None):
        d = {
            "gamma": self.gamma,
            "tau": self.tau,
            "delta": self.delta,
            "delta_factor": self.delta_factor,
            "penalty": self.penalty,
            "window": self.window,
        }
        if kind == "str":
            return "\n".join(f"{k:<13}: {v}" for k, v in d.items())
        return d

    def wp_c_print(self):
        dtw_cc.wps_print(self._wp, len(self.series1), len(self.series2), window=self.window)

    def wp_c_print_compact(self):
        dtw_cc.wps_print_compact(self._wp, len(self.series1), len(self.series2), window=self.window)



# -*- coding: UTF-8 -*-
"""
dtaidistance.subsequence.subsequencesearch
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.3.0 or higher)

DTW-based subsequence matching.

:author: Wannes Meert
:copyright: Copyright 2021-2023 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging

from .. import dtw  # import warping_paths, warping_paths_fast, best_path, warping_paths_affinity, distance
from .. import dtw_ndim
from .. import util_numpy
from .. import util


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    import numpy.ma as ma
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    ma = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None


def subsequence_search(query, series, dists_options=None, use_lb=True,
                       max_dist=None, max_value=None, use_c=None):
    """See SubsequenceSearch.

    :param query: Time series to search for
    :param series: Iterator over time series to perform search on.
            This can be for example windows over a long time series.
    :param dists_options: Options passed on to `dtw.distance`
    :param use_lb: Use lowerbounds to early abandon options
    :param max_dist: Ignore DTW distances larger than this value
    :param max_value: Ignore normalized DTW distances larger than this value
    :param use_c: Use fast C implementation if available
    :return: SubsequenceSearch object
    """
    ss = SubsequenceSearch(query, series, dists_options=dists_options, use_lb=use_lb,
                           max_dist=max_dist, max_value=max_value, use_c=use_c)
    return ss


class SSMatch:
    """Found match by SubsequenceSearch.

    The match is identified by the idx property, which is the index of the matched
    series in the original list of series. The distance property returns the DTW
    distance between the query and the series at index idx.
    """
    def __init__(self, kidx, ss):
        self.kidx = kidx
        self.ss = ss

    @property
    def distance(self):
        """DTW distance."""
        return self.ss.kbest_distances[self.kidx][0]

    @property
    def value(self):
        """Normalized DTW distance."""
        return self.distance / len(self.ss.query)

    @property
    def idx(self):
        return self.ss.kbest_distances[self.kidx][1]

    def __str__(self):
        return f'SSMatch({self.idx})'

    def __repr__(self):
        return self.__str__()


class SSMatches:
    def __init__(self, ss, k=None):
        """Iterator over matches.

        :param ss: The SubsequenceSearch object
        :param k: Optional a k. This overrules the ss.k value.
            Useful if a smaller k is asked to iterate over than has been stored.
        """
        self.ss = ss
        self.k = k
        if self.ss.kbest_distances is None:
            self.k = 0
        elif self.k is None or self.k > self.ss.k:
            self.k = self.ss.k
        if self.k is None:
            self.k = len(self.ss.kbest_distances)

    def __getitem__(self, key):
        if isinstance(key, slice):
            start = 0 if key.start is None else key.start
            return [SSMatch(kip+start, self.ss) for kip, (v, i) in
                    enumerate(self.ss.kbest_distances[key])]
        return SSMatch(key, self.ss)

    def __iter__(self):
        for ki, (v, i) in enumerate(self.ss.kbest_distances[:self.k]):
            yield SSMatch(ki, self.ss)

    def __len__(self):
        return self.k

    def __str__(self):
        if self.k > 10:
            return '[' + ', '.join(str(m) for m in self[:5]) + ' ... ' +\
                   ', '.join(str(m) for m in self[-5:]) + ']'
        return '[' + ', '.join(str(m) for m in self) + ']'


class SubsequenceSearch:
    def __init__(self, query, s, dists_options=None, use_lb=True, keep_all_distances=False,
                 max_dist=None, max_value=None, use_c=None, use_ndim=None):
        """Search the best matching (subsequence) time series compared to a given time series.

        :param query: Time series to search for
        :param s: Iterator over time series to perform search on.
            This can be for example windows over a long time series.
        :param dists_options: Options passed on to `dtw.distance`
        :param use_lb: Use lowerbounds to early abandon options
        :param max_dist: Ignore DTW distances larger than this value
            if max_dist is also given in dists_options, then the one in dists_options is ignored
            if both max_dist and max_value are given, the smallest is used
        :param max_value: Ignore normalized DTW distances larger than this value
        """
        self.query = query
        if use_ndim is None:
            self.use_ndim = (util.detect_ndim(query) > 1)
        else:
            self.use_ndim = use_ndim
        self.s = s
        # If keep_all_distances is true, store all. Can take up quite some memory.
        self.distances = None
        # Keep track of the k-best distances
        self.kbest_distances = None
        self.lbs = None
        self.k = None
        self.dists_options = {} if dists_options is None else dists_options
        if max_dist is None:
            self.max_dist = self.dists_options.get('max_dist', np.inf)
        else:
            self.max_dist = max_dist
        if max_value is not None:
            self.max_dist = min(self.max_dist, max_value * len(self.query))
        self.dists_options['max_dist'] = self.max_dist
        if use_c is not None:
            self.dists_options['use_c'] = use_c
        self.use_lb = use_lb

        self.keep_all_distances = keep_all_distances
        # if self.use_lb and not self.keep_all_distances:
        #     raise ValueError("If use_lb is true, then keep_all_distances should also be true.")

    def reset(self):
        self.distances = None
        self.kbest_distances = None
        self.lbs = None

    # def compute_lbs(self):
    #     self.lbs = np.zeros((len(self.s),))
    #     for idx, series in enumerate(self.s):
    #         self.lbs[idx] = dtw.lb_keogh(self.query, series, **self.dists_options)

    def align_fast(self, k=None):
        use_c = self.dists_options['use_c']
        self.dists_options['use_c'] = True
        result = self.align(k=k)
        self.dists_options['use_c'] = use_c
        return result

    def align(self, k=None):
        if k is not None and self.k is not None and k <= self.k and self.kbest_distances is not None:
            return self.kbest_distances[:k]
        if self.use_ndim:
            distance = dtw_ndim.distance
            lb_keogh = None
            if self.use_lb:
                self.use_lb = False
                logger.warning('The setting use_lb is ignored for multivariate series.')
        else:
            distance = dtw.distance
            lb_keogh = dtw.lb_keogh
        if k is None or self.keep_all_distances:
            self.distances = np.zeros((len(self.s),))
            # if self.use_lb:
            #     self.compute_lbs()
        import heapq
        h = [(-np.inf, -1)]
        max_dist = self.max_dist
        for idx, series in enumerate(self.s):
            if self.use_lb:
                lb = lb_keogh(self.query, series, **self.dists_options)
                if lb > max_dist:
                    continue
            dist = distance(self.query, series, **self.dists_options)
            if k is not None:
                if len(h) < k:
                    if not np.isinf(dist) and dist <= max_dist:
                        heapq.heappush(h, (-dist, idx))
                        max_dist = min(max_dist, -h[0][0])
                else:
                    if not np.isinf(dist) and dist <= max_dist:
                        heapq.heappushpop(h, (-dist, idx))
                        max_dist = min(max_dist, -h[0][0])
                self.dists_options['max_dist'] = max_dist
            if self.keep_all_distances or k is None:
                self.distances[idx] = dist
        if k is not None:
            # hh = np.array([-v for v, _ in h])
            # self.kbest_distances = [(-h[i][0], h[i][1]) for i in np.argsort(hh)]
            self.kbest_distances = sorted((-v, i) for v, i in h if i != -1)
        else:
            self.kbest_distances = [(self.distances[i], i) for i in np.argsort(self.distances)]

        self.k = k
        return self.kbest_distances

    def get_ith_value(self, i):
        """Return the i-th value from the k-best values.

        :param i: Return i-th best value (i < k)
        :return: (distance, index)
        """
        if self.kbest_distances is None or self.k is None:
            raise ValueError('Align should be called before asking for the i-th value.')
        if i > self.k:
            raise ValueError('The i-th value is not available, i={}>k={}'.format(i, self.k))
        return self.kbest_distances[i]

    def best_match_fast(self):
        self.dists_options['use_c'] = True
        return self.best_match()

    def best_match(self):
        self.align(k=1)
        # _value, best_idx = self.kbest_distances[0]
        return SSMatch(0, self)

    def kbest_matches_fast(self, k=1):
        use_c = self.dists_options.get('use_c', None)
        self.dists_options['use_c'] = True
        result = self.kbest_matches(k=k)
        if use_c is None:
            del self.dists_options['use_c']
        else:
            self.dists_options['use_c'] = use_c
        return result

    def kbest_matches(self, k=1):
        """Return the k best matches.

        It is recommended to set k to a value, and not None.
        If k is set to None, all comparisons are kept and returned. Also no early
        stopping is applied in case k is None.

        :param k: Number of best matches to return (default is 1)
        :return: List of SSMatch objects
        """
        self.align(k=k)
        # if k is None:
        #     return [SSMatch(best_idx, self) for best_idx in range(len(self.distances))]
        # if self.keep_all_distances:
        #     best_idxs = np.argpartition(self.distances, k)
            # return [SSMatch(best_idx, self) for best_idx in best_idxs[:k]]
        # distances = reversed(sorted(self.h))
        # return [SSMatch(best_idx, self) for dist, best_idx in distances]
        return SSMatches(self)



# -*- coding: UTF-8 -*-
"""
dtaidistance.preprocessing
~~~~~~~~~~~~~~~~~~~~~~~~~~

Preprocessing time series.

:author: Wannes Meert
:copyright: Copyright 2021-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
from .exceptions import NumpyException, ScipyException


def differencing(series, smooth=None):
    """Differencing series.

    :param series: Time series (must be numpy compatible)
    :param smooth: Smooth the series by removing the `smooth` percentage ([0-1])
        of highest frequency.
    :return: Differenced Numpy array
    """
    try:
        import numpy as np
    except ImportError:
        raise NumpyException("Differencing requires Numpy")
    try:
        from scipy import signal
    except ImportError:
        raise ScipyException("Differencing requires Scipy")
    if isinstance(series, np.ndarray):
        if len(series.shape) == 1:
            axis = 0
        else:
            axis = 1
    series = np.diff(series, n=1, axis=axis)
    if smooth is not None:
        fs = 100  # sample rate, Hz
        cutoff = fs * smooth  # cut off frequency, Hz
        nyq = 0.5 * fs  # Nyquist frequency
        b, a = signal.butter(2, cutoff / nyq, btype='low', analog=False, output='ba')
        try:
            series = signal.filtfilt(b, a, series, axis=axis)
        except ValueError as exc:
            raise ValueError("Cannot apply smoothing, "
                             "see the Scipy exception above to solve the problem "
                             "or disable smoothing by setting smooth to None") from exc
    return series


def znormal(series):
    """Z-normalize the time series.

    :param series: Time series (must be a numpy compatible)
    :return: Z-normalized Numpy array
    """
    try:
        import numpy as np
    except ImportError:
        raise NumpyException("Z normalization requires Numpy")
    series = np.array(series)
    series = (series - series.mean(axis=1)[:, None]) / series.std(axis=1)[:, None]
    return series



try:
    import numpy as np
except ImportError:
    np = None


def distance_to_similarity(D, r=None, method='exponential', return_params=False):
    """Transform a distance matrix to a similarity matrix.

    The available methods are:
    - Exponential: e^(-D / r)
      r is max(D) if not given
    - Gaussian: e^(-D^2 / r^2)
      r is max(D) if not given
    - Reciprocal: 1 / (r + D)
      r is 1 if not given
    - Reverse: r - D
      r is min(D) + max(D) if not given

    All of these methods are monotonically decreasing transformations. The order of the
    distances thus remains unchanged (only the direction).

    Example usage::

        dist_matrix = dtw.distance_matrix(series)
        sim_matrix = distance_to_similarity(dist_matrix)


    :param D: The distance matrix
    :param r: A scaling or smoothing parameter.
    :param method: One of 'exponential', 'gaussian', 'reciprocal', 'reverse'
    :return: Similarity matrix S
    """
    method = method.lower()
    if method == 'exponential':
        if r is None:
            r = np.max(D)
        S = np.exp(-D / r)
    elif method == 'gaussian':
        if r is None:
            r = np.max(D)
        S = np.exp(-np.power(D, 2) / r**2)
    elif method == 'reciprocal':
        if r is None:
            r = 1
        S = 1 / (r + D)
    elif method == 'reverse':
        if r is None:
            r = np.min(D) + np.max(D)
        S = (r - D) / r
    else:
        raise ValueError("method={} is not supported".format(method))
    if return_params:
        return S, r
    else:
        return S


def squash(X, r=None, base=None, x0=None, method="logistic", return_params=False):
    """Squash a function monotonically to a range between 0 and 1.

    The available methods are:
    - Logistic: 1 / (1 + e^(-(X-x0) / r)
    - Gaussian: e^(-(X-x0)^2 / r^2)

    Example usage::

        dist_matrix = dtw.distance_matrix(series)
        dist_matrix_sq = squash(dist_matrix)


    Based on:
    Vercruyssen, V., Meert, W., Verbruggen, G., Maes, K., Baumer, R., & Davis, J.
    (2018). Semi-supervised anomaly detection with an application to water analytics.
    In 2018 IEEE international conference on data mining (ICDM) (Vol. 2018, pp. 527-536)

    :param X: Distances values
    :param r: The slope of the squashing (see the formula above)
    :param x0: The midpoint of the squashing (see the formula above)
    :param method: The choice of sqaush function: logistic or gaussian
    :param return_params: Also return the used values for r and X0
    """
    result = None
    if method == "gaussian":
        x0 = 0  # not supported for gaussian
        if r is None:
            r = 1
        if base is None:
            result = 1 - np.exp(-np.power(X - x0, 2) / r**2)
        else:
            result = 1 - np.power(base, -np.power(X - x0, 2) / r**2)
    elif method == "logistic":
        if x0 is None:
            x0 = np.mean(X)
        if r is None:
            r = x0 / 6
        if base is None:
            result = 1 / (1 + np.exp(-(X - x0) / r))
        else:
            result = 1 / (1 + np.power(base, -(X - x0) / r))
    else:
        raise ValueError("Unknown value for method")
    if return_params:
        return result, r, x0
    else:
        return result



# -*- coding: UTF-8 -*-
"""
dtaidistance.subsequence.subsequencealignment
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.3.0 or higher)

DTW-based subsequence matching.

:author: Wannes Meert
:copyright: Copyright 2021-2023 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging

from .. import dtw  # import warping_paths, warping_paths_fast, best_path, warping_paths_affinity, distance
from .. import dtw_ndim
from .. import util_numpy
from .. import util


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    import numpy.ma as ma
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    ma = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None


def subsequence_alignment(query, series, use_c=False):
    """See SubsequenceAligment.

    :param query:
    :param series:
    :return:
    """
    sa = SubsequenceAlignment(query, series, use_c=use_c)
    sa.align()
    return sa


class SAMatch:
    def __init__(self, idx, alignment):
        """SubsequenceAlignment match"""
        self.idx = idx
        self.alignment = alignment

    @property
    def value(self):
        """Normalized DTW distance of match.

        Normalization is the DTW distance divided by the query length.
        """
        return self.alignment.matching[self.idx]

    @property
    def distance(self):
        """DTW distance of match.

        This value is dependent on the length of the query. Use the value
        property when comparing queries of different lengths.
        """
        return self.value * len(self.alignment.query)

    @property
    def segment(self):
        """Matched segment in series."""
        start = self.alignment.matching_function_startpoint(self.idx)
        end = self.alignment.matching_function_endpoint(self.idx)
        return [start, end]

    @property
    def path(self):
        """Matched path in series"""
        return self.alignment.matching_function_bestpath(self.idx)

    def __str__(self):
        return f'SAMatch({self.idx})'

    def __repr__(self):
        return self.__str__()


class SubsequenceAlignment:
    def __init__(self, query, series, penalty=0.1, use_c=False):
        """Subsequence alignment using DTW.
        Find where the query occurs in the series.

        Based on Fundamentals of Music Processing, Meinard Müller, Springer, 2015.

        Example::

            query = np.array([1., 2, 0])
            series = np.array([1., 0, 1, 2, 1, 0, 2, 0, 3, 0, 0])
            sa = subsequence_search(query, series)
            mf = sa.matching_function()
            sa.kbest_matches(k=2)


        :param query: Subsequence to search for
        :param series: Long sequence in which to search
        :param penalty: Penalty for non-diagonal matching
        :param use_c: Use the C-based DTW function if available
        """
        self.query = query
        self.series = series
        self.penalty = penalty
        self.paths = None
        self.matching = None
        self.use_c = use_c

    def reset(self):
        self.matching = None

    def align(self):
        if self.matching is not None:
            return
        psi = [0, 0, len(self.series), len(self.series)]
        if np is not None and isinstance(self.series, np.ndarray) and len(self.series.shape) > 1:
            if not self.use_c:
                _, self.paths = dtw_ndim.warping_paths(self.query, self.series, penalty=self.penalty, psi=psi,
                                                       psi_neg=False)
            else:
                _, self.paths = dtw_ndim.warping_paths_fast(self.query, self.series, penalty=self.penalty, psi=psi,
                                                            compact=False, psi_neg=False)
        else:
            if not self.use_c:
                _, self.paths = dtw.warping_paths(self.query, self.series, penalty=self.penalty, psi=psi,
                                                  psi_neg=False)
            else:
                _, self.paths = dtw.warping_paths_fast(self.query, self.series, penalty=self.penalty, psi=psi,
                                                       compact=False, psi_neg=False)
        self._compute_matching()

    def align_fast(self):
        use_c = self.use_c
        self.use_c = True
        result = self.align()
        self.use_c = use_c
        return result

    def _compute_matching(self):
        matching = self.paths[-1, :]
        if len(matching) > len(self.series):
            matching = matching[-len(self.series):]
        self.matching = np.array(matching) / len(self.query)

    def warping_paths(self):
        """Get matrix with all warping paths.

        If the aligmnent was computed using a compact, the paths are first copied into a full
        warping paths matrix.

        :return: Numpy matrix of size (len(query)+1) * (len(series)+1)
        """
        return self.paths

    def matching_function(self):
        """The matching score for each end-point of a possible match."""
        return self.matching

    def get_match(self, idx):
        return SAMatch(idx, self)

    def best_match_fast(self):
        use_c = self.use_c
        self.use_c = True
        result = self.best_match()
        self.use_c = use_c
        return result

    def best_match(self):
        best_idx = np.argmin(self.matching)
        return self.get_match(best_idx)

    def kbest_matches_fast(self, k=1, overlap=0):
        use_c = self.use_c
        self.use_c = True
        result = self.kbest_matches(k=k, overlap=overlap)
        self.use_c = use_c
        return result

    def kbest_matches(self, k=1, overlap=0):
        """Yields the next best match. Stops at k matches (use None for all matches).

        :param k: Number of matches to yield. None is all matches.
        :param overlap: Matches cannot overlap unless overlap > 0.
        :return: Yield an SAMatch object
        """
        self.align()
        matching = np.array(self.matching)
        maxv = np.ceil(np.max(matching) + 1)
        matching[:min(len(self.query) - 1, overlap)] = maxv
        ki = 0
        while k is None or ki < k:
            best_idx = np.argmin(matching)
            if best_idx == 0 or np.isinf(matching[best_idx]) or matching[best_idx] == maxv:
                # No more matches found
                break
            match = self.get_match(best_idx)
            b, e = match.segment
            cur_overlap = min(overlap, e - b - 1)
            mb, me = best_idx + 1 - (e - b) + cur_overlap, best_idx + 1
            if np.isinf(np.max(matching[mb:me])):
                # No overlapping matches
                matching[best_idx] = maxv
                continue
            matching[mb:me] = np.inf
            ki += 1
            yield match

    def matching_function_segment(self, idx):
        """Matched segment in series."""
        start = self.matching_function_startpoint(idx)
        end = self.matching_function_endpoint(idx)
        return [start, end]

    def matching_function_endpoint(self, idx):
        """Index in series for end of match in matching function at idx.

        :param idx: Index in matching function
        :return: Index in series
        """
        if len(self.matching) == len(self.series):
            return idx
        diff = len(self.series) - len(self.matching)
        return idx + diff

    def matching_function_startpoint(self, idx):
        """Index in series for start of match in matching function at idx.

        :param idx: Index in matching function
        :return: Index in series
        """
        real_idx = idx + 1
        path = dtw.best_path(self.paths, col=real_idx)
        start_idx = path[0][1]
        return start_idx

    def matching_function_bestpath(self, idx):
        """Indices in series for best path for match in matching function at idx.

        :param idx: Index in matching function
        :return: List of (row, col)
        """
        real_idx = idx + 1
        path = dtw.best_path(self.paths, col=real_idx)
        return path



# -*- coding: UTF-8 -*-
"""
dtaidistance.innerdistance
~~~~~~~~~~~~~~~~~~~~~~~~~~

Inner distances for DTW and ED

:author: Wannes Meert
:copyright: Copyright 2023 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import math
import logging

from . import util
from . import util_numpy

try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
    DTYPE = np.double
    argmin = np.argmin
    argmax = np.argmax
    array_min = np.min
    array_max = np.max
except ImportError:
    np = None
    argmin = util.argmin
    argmax = util.argmax
    array_min = min
    array_max = max


logger = logging.getLogger("be.kuleuven.dtai.distance")
default = 'squared euclidean'


class SquaredEuclidean:

    @staticmethod
    def inner_dist(x, y):
        return (x - y) ** 2

    @staticmethod
    def result(x):
        return math.sqrt(x)


class SquaredEuclideanNdim:

    @staticmethod
    def inner_dist(x, y):
        return np.sum((x - y) ** 2)

    @staticmethod
    def result(x):
        return np.sqrt(x)


class Euclidean:

    @staticmethod
    def inner_dist(x, y):
        return abs(x - y)

    @staticmethod
    def result(x):
        return x


class EuclideanNdim:

    @staticmethod
    def inner_dist(x, y):
        return np.sqrt(np.sum(np.power(x - y, 2)))

    @staticmethod
    def result(x):
        return x


class CustomInnerDist:

    @staticmethod
    def inner_dist(x, y):
        """The distance between two points in the series.

        For example, for default DTW this would be the Squared Euclidean
        distance: (a-b)**2.
        """
        raise Exception("Function not defined")

    @staticmethod
    def result(x):
        """The transformation applied to the sum of all inner distances.

        For example, for default DTW, which uses Squared Euclidean, this
        would be: sqrt(d). Because d = (a_0-b_0)**2 + (a_1-b_1)**2 ...
        """
        raise Exception("Function not defined")


def inner_dist_fns(inner_dist="squared euclidean", use_ndim=False):
    use_cls = None
    if inner_dist == "squared euclidean":
        if use_ndim:
            use_cls = SquaredEuclideanNdim
        else:
            use_cls = SquaredEuclidean
    elif inner_dist == "euclidean":
        if use_ndim:
            use_cls = EuclideanNdim
        else:
            use_cls = Euclidean
    elif hasattr(inner_dist, 'inner_dist') and hasattr(inner_dist, 'result'):
        use_cls = inner_dist
    else:
        raise AttributeError("Unknown value for argument inner_dist")
    return use_cls.inner_dist, use_cls.result


def to_c(inner_dist):
    if inner_dist == 'squared euclidean':
        return 0
    elif inner_dist == 'euclidean':
        return 1
    elif hasattr(inner_dist, 'inner_dist') and hasattr(inner_dist, 'result'):
        raise AttributeError('Custom inner distance functions are not supported for the fast C implementation')
    else:
        raise AttributeError('Unknown inner_dist: {}'.format(inner_dist))




{%- set s = " " %}
def distance_matrix{{suffix}}(cur,{{s}}
        {%- if "ndim" in suffix -%}
        int ndim,{{s}}
        {%- endif -%}
        block=None, **kwargs):
    """Compute a distance matrix between all sequences given in `cur`.
    This method calls a pure c implementation of the dtw computation that
    avoids the GIL.

    Assumes C-contiguous arrays.

    :param cur: DTWSeriesMatrix or DTWSeriesPointers
    :param block: see DTWBlock
    :param kwargs: Settings (see DTWSettings)
    :return: The distance matrix as a list representing the triangular matrix.
    """
    cdef DTWSeriesMatrix matrix
    {%- if "ndim" in suffix %}
    cdef DTWSeriesMatrixNDim matrixnd
    {%- endif %}
    cdef DTWSeriesPointers ptrs
    cdef Py_ssize_t length = 0
    cdef Py_ssize_t block_rb=0
    cdef Py_ssize_t block_re=0
    cdef Py_ssize_t block_cb=0
    cdef Py_ssize_t block_ce=0
    cdef Py_ssize_t ri = 0
    if block is not None and block != 0.0:
        block_rb = block[0][0]
        block_re = block[0][1]
        block_cb = block[1][0]
        block_ce = block[1][1]

    settings = DTWSettings(**kwargs)
    cdef DTWBlock dtwblock = DTWBlock(rb=block_rb, re=block_re, cb=block_cb, ce=block_ce)
    if block is not None and block != 0.0 and len(block) > 2 and block[2] is False:
        dtwblock.triu_set(False)
    length = distance_matrix_length(dtwblock, len(cur))

    # Correct block
    if dtwblock.re == 0:
        dtwblock.re_set(len(cur))
    if dtwblock.ce == 0:
        dtwblock.ce_set(len(cur))

    cdef array.array dists = array.array('d')
    array.resize(dists, length)

    if isinstance(cur, DTWSeriesMatrix){{s}}
        {%- if "ndim" in suffix %}or isinstance(cur, DTWSeriesMatrixNDim) {% endif -%}
        or isinstance(cur, DTWSeriesPointers):
        pass
    elif cur.__class__.__name__ == "SeriesContainer":
        cur = cur.c_data_compat()
    else:
        cur = dtw_series_from_data(cur{%- if "ndim" in suffix %}, force_pointers=True{% endif -%})

    if isinstance(cur, DTWSeriesPointers):
        ptrs = cur
        {%- if "ndim" in suffix %}
        dtaidistancec_dtw.dtw_distances_ndim_ptrs(
            ptrs._ptrs, ptrs._nb_ptrs, ptrs._lengths, ndim,
            dists.data.as_{{seq_t}}s, &dtwblock._block, &settings._settings)
        {%- else %}
        dtaidistancec_dtw.dtw_distances_ptrs(
            ptrs._ptrs, ptrs._nb_ptrs, ptrs._lengths,
            dists.data.as_{{seq_t}}s, &dtwblock._block, &settings._settings)
        {%- endif %}
    elif isinstance(cur, DTWSeriesMatrix):
        {%- if "ndim" in suffix %}
        # This is not a n-dimensional case ?
        {%- endif %}
        matrix = cur
        dtaidistancec_dtw.dtw_distances_matrix(
            &matrix._data[0,0], matrix.nb_rows, matrix.nb_cols,
            dists.data.as_{{seq_t}}s, &dtwblock._block, &settings._settings)
    {%- if "ndim" in suffix %}
    elif isinstance(cur, DTWSeriesMatrixNDim):
        matrixnd = cur
        dtaidistancec_dtw.dtw_distances_ndim_matrix(
            &matrixnd._data[0,0,0], matrixnd.nb_rows, matrixnd.nb_cols, ndim,
            dists.data.as_{{seq_t}}s, &dtwblock._block, &settings._settings)
    {%- endif %}
    else:
        raise Exception("Unknown series container")

    return dists



# -*- coding: UTF-8 -*-
"""
dtaidistance.clustering.kmeans
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

(requires version 2.2.0 or higher)

Time series clustering using k-means and Barycenter averages.

:author: Wannes Meert
:copyright: Copyright 2020-2022 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import random
import math
import multiprocessing as mp


logger = logging.getLogger("be.kuleuven.dtai.distance")

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

dtw_cc = None
try:
    from .. import dtw_cc
except ImportError:
    logger.debug('DTAIDistance C library not available')
    dtw_cc = None


try:
    import numpy as np
except ImportError:
    np = None

from ..dtw import distance, distance_matrix_fast, distance_matrix
from  .medoids import KMedoids
from ..util import SeriesContainer
from ..exceptions import NumpyException, PyClusteringException, MatplotlibException
from .visualization import prepare_plot_options
from .medoids import Medoids
from ..dtw_barycenter import dba_loop
from .. import dtw, dtw_ndim


def _distance_with_params(t):
    series, avgs, dists_options = t
    min_i, min_d = -1, float('inf')
    for i, avg in enumerate(avgs):
        d = distance(series, avg, **dists_options)
        if d < min_d:
            min_d, min_i = d, i
    return min_i, min_d


def _distance_ndim_with_params(t):
    series, avgs, dists_options = t
    min_i, min_d = -1, float('inf')
    for i, avg in enumerate(avgs):
        d = dtw_ndim.distance(series, avg, **dists_options)
        if d < min_d:
            min_d, min_i = d, i
    return min_i, min_d


def _distance_c_with_params(t):
    series, means, dists_options = t
    min_i, min_d = -1, float('inf')
    for i, mean in enumerate(means):
        d = dtw_cc.distance(series, mean, **dists_options)
        if d < min_d:
            min_d, min_i = d, i
    return min_i, min_d


def _distance_ndim_c_with_params(t):
    series, means, dists_options = t
    min_i, min_d = -1, float('inf')
    for i, mean in enumerate(means):
        d = dtw_cc.distance_ndim(series, mean, **dists_options)
        if d < min_d:
            min_d, min_i = d, i
    return min_i, min_d


def _dba_loop_with_params(t):
    series, c, mask, max_it, thr, use_c, nb_prob_samples = t
    return dba_loop(series, c=c, mask=mask, max_it=max_it, thr=thr, use_c=use_c,
                    nb_prob_samples=nb_prob_samples)


class KMeans(Medoids):
    def __init__(self, k, max_it=10, max_dba_it=10, thr=0.0001, drop_stddev=None,
                 nb_prob_samples=None,
                 dists_options=None, show_progress=True,
                 initialize_with_kmedoids=False, initialize_with_kmeanspp=True,
                 initialize_sample_size=None):
        """K-means clustering algorithm for time series using Dynamic Barycenter
        Averaging.

        :param k: Number of components
        :param max_it: Maximal interations for K-means
        :param max_dba_it: Maximal iterations for the Dynamic Barycenter Averaging.
        :param thr: Convergence is achieved if the averaging iterations differ less
            then this threshold
        :param drop_stddev: When computing the average series per cluster, ignore
            the instances that are further away than stddev*drop_stddev from the
            prototype (this is a gradual effect, the algorithm starts with drop_stddev
            is 3).
            This is related to robust k-means approaches that use trimming functions.
        :param nb_prob_samples: Probabilistically sample best path this number of times.
        :param dists_options:
        :param show_progress:
        :param initialize_with_kmedoids: Cluster a sample of the dataset first using
            K-medoids.
        :param initialize_with_kmeanspp: Use k-means++
        :param initialize_sample_size: How many samples to use for initialization with K-medoids or K-means++.
            Defaults are k*20 for K-medoid and 2+log(k) for k-means++.
        """
        if dists_options is None:
            dists_options = {}
        self.means = [None] * k
        self.max_it = max_it
        self.max_dba_it = max_dba_it
        self.thr = thr
        self.drop_stddev = drop_stddev

        self.initialize_with_kmeanspp = initialize_with_kmeanspp
        self.initialize_with_kmedoids = initialize_with_kmedoids
        self.initialize_sample_size = initialize_sample_size
        self.nb_prob_samples = nb_prob_samples
        super().__init__(None, dists_options, k, show_progress)

    def kmedoids_centers(self, series, use_c=False):
        logger.debug('Start K-medoid initialization ... ')
        if self.initialize_sample_size is None:
            sample_size = min(self.k * 20, len(self.series))
        else:
            sample_size = min(self.initialize_sample_size, len(self.series))
        indices = np.random.choice(range(0, len(self.series)), sample_size, replace=False)
        sample = self.series[indices, :].copy()
        if use_c:
            fn_dm = distance_matrix_fast
        else:
            fn_dm = distance_matrix
        model = KMedoids(fn_dm, {**self.dists_options, **{'compact': False}}, k=self.k)
        cluster_idx = model.fit(sample)
        means = [self.series[idx] for idx in cluster_idx.keys()]
        logger.debug('... Done')
        return means

    def kmeansplusplus_centers(self, series, use_c=False):
        """Better initialization for K-Means.

            Arthur, D., and S. Vassilvitskii. "k-means++: the, advantages of careful seeding.
            In, SODA’07: Proceedings of the." eighteenth annual ACM-SIAM symposium on Discrete, algorithms.

        Procedure (in paper):

        - 1a. Choose an initial center c_1 uniformly at random from X.
        - 1b. Choose the next center c_i , selecting c_i = x′∈X with probability D(x')^2/sum(D(x)^2, x∈X).
        - 1c. Repeat Step 1b until we have chosen a total of k centers.
        - (2-4. Proceed as with the standard k-means algorithm.)

        Extension (in conclusion):

        - Also, experiments showed that k-means++ generally performed better if it selected several new centers
          during each iteration, and then greedily chose the one that decreased φ as much as possible.

        Detail (in original code):

        - numLocalTries==2+log(k)

        :param series:
        :param use_c:
        :return:
        """
        if np is None:
            raise NumpyException("Numpy is required for the KMeans.kmeansplusplus_centers method.")
        logger.debug('Start K-means++ initialization ... ')
        ndim = self.series.detected_ndim
        print(self.dists_options)
        print(dtw_ndim.distance(series[0], series[5], **self.dists_options))
        if use_c:
            if ndim == 1:
                fn = distance_matrix_fast
            else:
                fn = dtw_ndim.distance_matrix_fast
        else:
            if ndim == 1:
                fn = distance_matrix
            else:
                fn = dtw_ndim.distance_matrix
        indices = []
        if self.initialize_sample_size is None:
            n_samples = min(2 + int(math.log(self.k)), len(series) - self.k)
        else:
            n_samples = self.initialize_sample_size
        dists = np.empty((n_samples, len(series)))

        # First center is chosen randomly
        idx = np.random.randint(0, len(series))
        min_dists = np.power(fn(series, block=((idx, idx + 1), (0, len(series)), False),
                                compact=True, **self.dists_options), 2)
        indices.append(idx)

        for k_idx in range(1, self.k):
            # Compute the distance between each series and the nearest center that has already been chosen.
            # (Choose one new series at random as a new center, using a weighted probability distribution)
            # Select several new centers and then greedily chose the one that decreases pot as much as possible
            sum_min_dists = np.sum(min_dists)
            if sum_min_dists == 0.0:
                logger.warning('There are only {} < k={} different series'.format(k_idx, self.k))
                weights = None
            else:
                weights = min_dists / sum_min_dists
            idx_cand = np.random.choice(len(min_dists), size=n_samples, replace=False, p=weights)
            for s_idx, idx in enumerate(idx_cand):
                dists[s_idx, :] = np.power(fn(series, block=((idx, idx + 1), (0, len(series)), False),
                                              compact=True, **self.dists_options), 2)
                np.minimum(dists[s_idx, :], min_dists, out=dists[s_idx, :])
            potentials = np.sum(dists, axis=1)
            best_pot_idx = np.argmin(potentials)
            idx = idx_cand[best_pot_idx]
            min_dists[:] = dists[best_pot_idx, :]
            indices.append(idx)

        means = [series[i] for i in indices]
        assert(len(means) == self.k)
        logger.debug('... Done')
        return means

    def fit_fast(self, series, monitor_distances=None):
        return self.fit(series, use_c=True, use_parallel=True, monitor_distances=monitor_distances)

    def fit(self, series, use_c=False, use_parallel=True, monitor_distances=None):
        """Perform K-means clustering.

        :param series: Container with series
        :param use_c: Use the C-library (only available if package is compiled)
        :param use_parallel: Use multipool for parallelization
        :param monitor_distances: This function is called with two arguments:
            (1) a list of (cluster, distance) for each instance;
            (2) a boolean indicating whether the clustering has been stopped or not.
            From this one can compute inertia or other metrics
            to monitor the clustering. If the boolean argument is true, this is the
            final assignment. If this function returns True, the clustering
            continues, if False is returned the clustering is stopped.
        :return: cluster indices, number of iterations
            If the number of iterations is equal to max_it, the clustering
            did not converge.
        """
        if np is None:
            raise NumpyException("Numpy is required for the KMeans.fit method.")
        self.series = SeriesContainer.wrap(series, support_ndim=True)
        ndim = self.series.detected_ndim
        mask = np.full((self.k, len(self.series)), False, dtype=bool)
        mask_new = np.full((self.k, len(self.series)), False, dtype=bool)
        means = [None] * self.k
        diff = 0
        performed_it = 1
        clusters = None
        if self.drop_stddev is not None:
            # Gradually tighten drop_stddev
            drop_stddev = max(self.drop_stddev, 4)
        else:
            drop_stddev = None

        if use_c:
            if ndim == 1:
                fn = _distance_c_with_params
            else:
                fn = _distance_ndim_c_with_params
        else:
            if ndim == 1:
                fn = _distance_with_params
            else:
                fn = _distance_ndim_with_params

        # Initialisations
        if self.initialize_with_kmeanspp:
            self.means = self.kmeansplusplus_centers(self.series, use_c=use_c)
        elif self.initialize_with_kmedoids:
            self.means = self.kmedoids_centers(self.series, use_c=use_c)
        else:
            indices = np.random.choice(range(0, len(self.series)), self.k, replace=False)
            self.means = [self.series[random.randint(0, len(self.series) - 1)] for _ki in indices]

        # Iterations
        it_nbs = range(self.max_it)
        if self.show_progress and tqdm is not None:
            it_nbs = tqdm(it_nbs)
        for it_nb in it_nbs:

            # Assignment step
            performed_it += 1
            if use_parallel:
                with mp.Pool() as p:
                    clusters_distances = p.map(fn, [(self.series[idx], self.means, self.dists_options) for idx in
                                                    range(len(self.series))])
            else:
                clusters_distances = list(map(fn, [(self.series[idx], self.means, self.dists_options) for idx in
                                                   range(len(self.series))]))
            if monitor_distances is not None:
                cont = monitor_distances(clusters_distances, False)
                if cont is False:
                    break
            clusters, distances = zip(*clusters_distances)
            distances = list(distances)

            best_medoid = [0]*self.k
            best_dist = [float('inf')]*self.k
            for idx, (cluster, distance) in enumerate(clusters_distances):
                if distance < best_dist[cluster]:
                    best_dist[cluster] = distance
                    best_medoid[cluster] = idx

            if self.drop_stddev is not None and self.drop_stddev != 0:
                logger.debug('drop_stddev = {}'.format(drop_stddev))
                stats = []
                max_value = []
                for ki in range(self.k):
                    stats.append([0, 0, 0])
                for (cluster, distance) in clusters_distances:
                    stats[cluster][0] += distance
                    stats[cluster][2] += 1
                for ki in range(self.k):
                    if stats[ki][2] == 0:
                        stats[ki][0] = 0
                    else:
                        stats[ki][0] /= stats[ki][2]
                for (cluster, distance) in clusters_distances:
                    stats[cluster][1] += (stats[cluster][0] - distance)**2
                for ki in range(self.k):
                    if stats[ki][2] == 0:
                        stats[ki][1] = 0
                    else:
                        stats[ki][1] = math.sqrt(stats[ki][1]/stats[ki][2])
                    max_value.append(stats[ki][0] + stats[ki][1]*drop_stddev)
                drop_stddev = (drop_stddev + self.drop_stddev) / 2
            else:
                max_value = None

            mask_new[:, :] = False
            cnt = [0] * self.k
            for idx, (cluster, distance) in enumerate(clusters_distances):
                if max_value is None or distance <= max_value[cluster]:
                    # Ignore the far away instances to compute new averages.
                    mask_new[cluster, idx] = True
                else:
                    cnt[cluster] += 1
            logger.debug('Ignored instances: {} / {} (max_value = {})'.format(cnt, len(clusters_distances), max_value))
            if (mask == mask_new).all():
                logger.info("Stopped after {} iterations, no change in cluster assignment".format(it_nb))
                break
            mask[:, :] = mask_new
            for ki in range(self.k):
                if not mask[ki, :].any():
                    idx = np.argmax(distances)
                    logger.debug('Empty cluster {}, assigned most dissimilar sequence {}={}'.format(ki, idx, distances[ki]))
                    mask[:, idx] = False
                    mask[ki, idx] = True
                    distances[idx] = 0

            # Update step
            diff = 0
            difflen = 0
            if use_parallel:
                with mp.Pool() as p:
                    means = p.map(_dba_loop_with_params,
                                  [(self.series, self.series[best_medoid[ki]], mask[ki, :],
                                    self.max_dba_it, self.thr, use_c, self.nb_prob_samples) for ki in range(self.k)])
            else:
                means = list(map(_dba_loop_with_params,
                             [(self.series, self.series[best_medoid[ki]], mask[ki, :],
                               self.max_dba_it, self.thr, use_c, self.nb_prob_samples) for ki in range(self.k)]))
            # for ki in range(self.k):
            #     means[ki] = dba_loop(self.series, c=None, mask=mask[:, ki], use_c=True)
            for ki in range(self.k):
                curlen = min(len(means[ki]), len(self.means[ki]))
                difflen += curlen
                if ndim == 1:
                    for a, b in zip(means[ki], self.means[ki]):
                        diff += abs(a - b)
                else:
                    for a, b in zip(means[ki], self.means[ki]):
                        diff += max(abs(a[d] - b[d]) for d in range(ndim))
                self.means[ki] = means[ki]
            diff /= difflen
            if diff <= self.thr:
                print("Stopped early after {} iterations, no change in means".format(it_nb))
                break

        # Final assignment
        if use_parallel:
            with mp.Pool() as p:
                clusters_distances = p.map(fn, [(self.series[idx], self.means, self.dists_options) for idx in
                                                range(len(self.series))])
        else:
            clusters_distances = list(map(fn, [(self.series[idx], self.means, self.dists_options) for idx in
                                               range(len(self.series))]))
        if monitor_distances is not None:
            monitor_distances(clusters_distances, True)
        clusters, distances = zip(*clusters_distances)

        # self.cluster_idx = {medoid: {inst for inst in instances}
        #                       for medoid, instances in zip(medoids, clusters)}
        self.cluster_idx = {ki: set() for ki in range(self.k)}
        for idx, cluster in enumerate(clusters):
            self.cluster_idx[cluster].add(idx)
        return self.cluster_idx, performed_it
